import ctypes
import sys
from typing import Any, Callable, Dict, Optional, Tuple, TYPE_CHECKING, TypeVar, Union
import weakref

from pyglet.media.drivers.pulse import lib_pulseaudio as pa
from pyglet.media.exceptions import MediaException
from pyglet.util import debug_print

if TYPE_CHECKING:
    from pyglet.media.codecs import AudioData, AudioFormat

T = TypeVar('T')
PulseAudioStreamSuccessCallback = Callable[[ctypes.POINTER(pa.pa_stream), int, Any], Any]
PulseAudioStreamRequestCallback = Callable[[ctypes.POINTER(pa.pa_stream), int, Any], Any]
PulseAudioStreamNotifyCallback = Callable[[ctypes.POINTER(pa.pa_stream), Any], Any]
PulseAudioContextSuccessCallback = Callable[[ctypes.POINTER(pa.pa_context), int, Any], Any]


_debug = debug_print('debug_media')


_UINT32_MAX = 0xFFFFFFFF
_SIZE_T_MAX = sys.maxsize*2 + 1
PA_INVALID_INDEX = _UINT32_MAX
PA_INVALID_WRITABLE_SIZE = _SIZE_T_MAX


def get_uint32_or_none(value: int) -> Optional[int]:
    if value == _UINT32_MAX:
        return None
    return value


def get_bool_or_none(value: int) -> Optional[bool]:
    if value < 0:
        return None
    elif value == 1:
        return True
    else:
        return False


def get_ascii_str_or_none(value: Optional[bytes]) -> Optional[str]:
    if value is not None:
        return value.decode('ascii')
    return None


class Proplist:
    def __init__(self, ini_data: Optional[Dict[str, Union[bytes, str]]] = None) -> None:
        self._pl = pa.pa_proplist_new()
        if not self._pl:
            raise PulseAudioException(0, 'Failed creating proplist.')
        if ini_data is not None:
            for k, v in ini_data:
                self[k] = v

    def __setitem__(self, k, v):
        if isinstance(v, bytes):
            r = pa.pa_proplist_set(self._pl, k.encode("utf-8"), v, len(v))
        else:
            r = pa.pa_proplist_sets(self._pl, k.encode("utf-8"), v.encode("utf-8"))
        if r != 0:
            raise PulseAudioException(0, 'Error setting proplist entry.')

    def __delitem__(self, k):
        if pa.pa_proplist_unset(k) != 0:
            raise PulseAudioException(0, 'Error unsetting proplist entry.')

    def delete(self) -> None:
        pa.pa_proplist_free(self._pl)
        self._pl = None


class PulseAudioException(MediaException):
    def __init__(self, error_code: int, message: str) -> None:
        super().__init__(message)
        self.error_code = error_code
        self.message = message

    def __str__(self):
        return f'{self.__class__.__name__}: [{self.error_code}] {self.message}'

    __repr__ = __str__


class _MainloopLock:
    def __init__(self, mainloop: 'PulseAudioMainloop') -> None:
        self.mainloop = mainloop

    def __enter__(self):
        self.mainloop.lock_()

    def __exit__(self, _exc_type, _ecx_value, _tb):
        self.mainloop.unlock()


class PulseAudioMainloop:
    def __init__(self) -> None:
        self._pa_threaded_mainloop = pa.pa_threaded_mainloop_new()
        self._pa_mainloop_vtab = pa.pa_threaded_mainloop_get_api(self._pa_threaded_mainloop)
        self.lock = _MainloopLock(self)

    def start(self) -> None:
        """Start running the mainloop."""
        result = pa.pa_threaded_mainloop_start(self._pa_threaded_mainloop)
        if result < 0:
            raise PulseAudioException(0, "Failed to start PulseAudio mainloop")
        assert _debug('PulseAudioMainloop: Started')

    def delete(self) -> None:
        """Clean up the mainloop."""
        if self._pa_threaded_mainloop is not None:
            assert _debug("Delete PulseAudioMainloop")
            pa.pa_threaded_mainloop_stop(self._pa_threaded_mainloop)
            pa.pa_threaded_mainloop_free(self._pa_threaded_mainloop)
            self._pa_threaded_mainloop = None
            self._pa_mainloop_vtab = None

    def lock_(self) -> None:
        """Lock the threaded mainloop against events.  Required for all
        calls into PA."""
        assert self._pa_threaded_mainloop is not None
        pa.pa_threaded_mainloop_lock(self._pa_threaded_mainloop)

    def unlock(self) -> None:
        """Unlock the mainloop thread."""
        assert self._pa_threaded_mainloop is not None
        pa.pa_threaded_mainloop_unlock(self._pa_threaded_mainloop)

    def signal(self) -> None:
        """Signal the mainloop thread to break from a wait."""
        assert self._pa_threaded_mainloop is not None
        pa.pa_threaded_mainloop_signal(self._pa_threaded_mainloop, 0)

    def wait(self) -> None:
        """Unlock and then Wait for a signal from the locked mainloop.
        It's important to note that the PA mainloop lock is reentrant, yet this method only
        releases one lock.
        Before returning, the lock is reacquired.
        """
        assert self._pa_threaded_mainloop is not None
        pa.pa_threaded_mainloop_wait(self._pa_threaded_mainloop)

    def create_context(self) -> 'PulseAudioContext':
        """Construct and return a new context in this mainloop.
        Will grab the lock.
        """
        assert self._pa_mainloop_vtab is not None
        app_name = self._get_app_name().encode("utf-8")
        with self.lock:
            return PulseAudioContext(self, app_name)

    def _get_app_name(self) -> str:
        """Get the application name as advertised to the pulseaudio server."""
        # TODO move app name into pyglet.app (also useful for OS X menu bar?).
        return sys.argv[0]


class PulseAudioMainloopChild:
    def __init__(self, mainloop: PulseAudioMainloop):
        assert mainloop is not None
        self.mainloop = mainloop


class PulseAudioContext(PulseAudioMainloopChild):
    """Basic object for a connection to a PulseAudio server."""
    _state_name = {pa.PA_CONTEXT_UNCONNECTED: 'Unconnected',
                   pa.PA_CONTEXT_CONNECTING: 'Connecting',
                   pa.PA_CONTEXT_AUTHORIZING: 'Authorizing',
                   pa.PA_CONTEXT_SETTING_NAME: 'Setting Name',
                   pa.PA_CONTEXT_READY: 'Ready',
                   pa.PA_CONTEXT_FAILED: 'Failed',
                   pa.PA_CONTEXT_TERMINATED: 'Terminated'}

    def __init__(self, mainloop: PulseAudioMainloop, name: bytes) -> None:
        super().__init__(mainloop)

        # TODO: Filling in stuff like language, IDs and icons is possible here
        # but gateways to get them down here don't really exist.
        # pl = Proplist({}); pl.delete()
        ctx = pa.pa_context_new_with_proplist(mainloop._pa_mainloop_vtab, name, None)
        self.check_ptr_not_null(ctx)

        self._pa_context = ctx
        self.state = None

        self._set_state_callback(self._state_callback)

    def delete(self) -> None:
        """Completely shut down pulseaudio client. Will lock."""
        if self._pa_context is not None:
            with self.mainloop.lock:
                assert _debug("PulseAudioContext.delete")
                if self.is_ready:
                    pa.pa_context_disconnect(self._pa_context)

                    while self.state is not None and not self.is_terminated:
                        self.mainloop.wait()

                self._set_state_callback(0)
                pa.pa_context_unref(self._pa_context)

            self._pa_context = None

    @property
    def is_ready(self) -> bool:
        return self.state == pa.PA_CONTEXT_READY

    @property
    def is_failed(self) -> bool:
        return self.state == pa.PA_CONTEXT_FAILED

    @property
    def is_terminated(self) -> bool:
        return self.state == pa.PA_CONTEXT_TERMINATED

    @property
    def server(self) -> Optional[str]:
        if self.is_ready:
            return get_ascii_str_or_none(pa.pa_context_get_server(self._pa_context))
        return None

    @property
    def protocol_version(self) -> Optional[str]:
        if self._pa_context is not None:
            return get_uint32_or_none(pa.pa_context_get_protocol_version(self._pa_context))
        return None

    @property
    def server_protocol_version(self) -> Optional[str]:
        if self._pa_context is not None:
            return get_uint32_or_none(pa.pa_context_get_server_protocol_version(self._pa_context))
        return None

    @property
    def is_local(self) -> Optional[bool]:
        if self._pa_context is not None:
            return get_bool_or_none(pa.pa_context_is_local(self._pa_context))
        return None

    def connect(self, server: Optional[bytes] = None) -> None:
        """Connect the context to a PulseAudio server.

        Will grab the mainloop lock.

        :Parameters:
            `server` : bytes
                Server to connect to, or ``None`` for the default local
                server (which may be spawned as a daemon if no server is
                found).
        """
        assert self._pa_context is not None
        self.state = None

        with self.mainloop.lock:
            self.check(
                pa.pa_context_connect(self._pa_context, server, 0, None)
            )
            while not self.is_failed and not self.is_ready:
                self.mainloop.wait()

            if self.is_failed:
                self.raise_error()

    def create_stream(self, audio_format: 'AudioFormat') -> 'PulseAudioStream':
        """
        Create a new audio stream.
        """
        assert self.is_ready

        return PulseAudioStream(self, audio_format)

    def set_input_volume(self, stream: 'PulseAudioStream', volume: float) -> 'PulseAudioOperation':
        """
        Set the volume for a stream.
        """
        cvolume = self._get_cvolume_from_linear(stream, volume)

        clump = PulseAudioContextSuccessCallbackLump(self)
        return PulseAudioOperation(
            clump,
            pa.pa_context_set_sink_input_volume(
                self._pa_context,
                stream.index,
                cvolume,
                clump.pa_callback,
                None,
            ),
        )

    def _get_cvolume_from_linear(self, stream: 'PulseAudioStream', volume: float) -> pa.pa_cvolume:
        cvolume = pa.pa_cvolume()
        volume = pa.pa_sw_volume_from_linear(volume)
        pa.pa_cvolume_set(cvolume, stream.get_sample_spec().channels, volume)
        return cvolume

    def _set_state_callback(
        self,
        py_callback: Optional[Callable[['PulseAudioContext', Any], Any]]
    ) -> None:
        if py_callback is None:
            self._pa_state_change_callback = None
        else:
            self._pa_state_change_callback = pa.pa_context_notify_cb_t(py_callback)
        pa.pa_context_set_state_callback(self._pa_context, self._pa_state_change_callback, None)

    def _state_callback(self, context: 'PulseAudioContext', _userdata) -> None:
        self.state = pa.pa_context_get_state(self._pa_context)
        assert _debug(f'PulseAudioContext: state changed to {self._state_name[self.state]}')
        self.mainloop.signal()

    def check(self, result: T) -> T:
        if result < 0:
            self.raise_error()
        return result

    def check_not_null(self, value: T) -> T:
        if value is None:
            self.raise_error()
        return value

    def check_ptr_not_null(self, value: T) -> T:
        if not value:
            self.raise_error()
        return value

    def raise_error(self) -> None:
        error = pa.pa_context_errno(self._pa_context)
        raise PulseAudioException(error, get_ascii_str_or_none(pa.pa_strerror(error)))


class PulseAudioStream(PulseAudioMainloopChild):
    """PulseAudio audio stream."""

    _state_name = {pa.PA_STREAM_UNCONNECTED: 'Unconnected',
                   pa.PA_STREAM_CREATING: 'Creating',
                   pa.PA_STREAM_READY: 'Ready',
                   pa.PA_STREAM_FAILED: 'Failed',
                   pa.PA_STREAM_TERMINATED: 'Terminated'}

    def __init__(self, context: PulseAudioContext, audio_format: 'AudioFormat') -> None:
        super().__init__(context.mainloop)

        self.state = None
        """The stream's state."""

        self.index = None
        """The stream's sink index."""

        self.context = weakref.ref(context)

        self._cb_write = pa.pa_stream_request_cb_t(0)
        self._cb_underflow = pa.pa_stream_notify_cb_t(0)
        self._cb_state = pa.pa_stream_notify_cb_t(self._state_callback)
        self._cb_moved = pa.pa_stream_notify_cb_t(self._moved_callback)

        self._pa_stream = pa.pa_stream_new_with_proplist(
            context._pa_context,
            f'{id(self):X}'.encode("utf-8"),
            self.create_sample_spec(audio_format),
            None,  # Default channel map
            None,  # No proplist to supply
        )
        context.check_not_null(self._pa_stream)

        pa.pa_stream_set_state_callback(self._pa_stream, self._cb_state, None)
        pa.pa_stream_set_moved_callback(self._pa_stream, self._cb_moved, None)
        self._refresh_state()

    def create_sample_spec(self, audio_format: 'AudioFormat') -> pa.pa_sample_spec:
        """
        Create a PulseAudio sample spec from pyglet audio format.
        """
        _FORMATS = {
            ('little', 8):  pa.PA_SAMPLE_U8,
            ('big', 8):     pa.PA_SAMPLE_U8,
            ('little', 16): pa.PA_SAMPLE_S16LE,
            ('big', 16):    pa.PA_SAMPLE_S16BE,
            ('little', 24): pa.PA_SAMPLE_S24LE,
            ('big', 24):    pa.PA_SAMPLE_S24BE,
        }
        fmt = (sys.byteorder, audio_format.sample_size) 
        if fmt not in _FORMATS:
            raise MediaException(f'Unsupported sample size/format: {fmt}')

        sample_spec = pa.pa_sample_spec()
        sample_spec.format = _FORMATS[fmt]
        sample_spec.rate = audio_format.sample_rate
        sample_spec.channels = audio_format.channels
        return sample_spec

    def delete(self) -> None:
        """If connected, disconnect, and delete the stream."""
        context = self.context()
        if context is None:
            assert _debug("No active context anymore. Cannot disconnect the stream")
            self._pa_stream = None
            return

        if self._pa_stream is None:
            assert _debug("No stream to delete.")
            return

        assert _debug("Delete PulseAudioStream")
        if not self.is_unconnected:
            assert _debug("PulseAudioStream: disconnecting")

            context.check(
                pa.pa_stream_disconnect(self._pa_stream)
            )
            while not (self.is_terminated or self.is_failed):
                self.mainloop.wait()

        self._disconnect_callbacks()
        pa.pa_stream_unref(self._pa_stream)
        self._pa_stream = None

    @property
    def is_unconnected(self) -> bool:
        return self.state == pa.PA_STREAM_UNCONNECTED

    @property
    def is_creating(self) -> bool:
        return self.state == pa.PA_STREAM_CREATING

    @property
    def is_ready(self) -> bool:
        return self.state == pa.PA_STREAM_READY

    @property
    def is_failed(self) -> bool:
        return self.state == pa.PA_STREAM_FAILED

    @property
    def is_terminated(self) -> bool:
        return self.state == pa.PA_STREAM_TERMINATED

    def get_writable_size(self) -> int:
        assert self._pa_stream is not None

        r = pa.pa_stream_writable_size(self._pa_stream)
        if r == PA_INVALID_WRITABLE_SIZE:
            self.context().raise_error()
        return r

    def is_corked(self) -> bool:
        assert self._pa_stream is not None

        r = pa.pa_stream_is_corked(self._pa_stream)
        self.context().check(r)
        return bool(r)

    def get_sample_spec(self) -> pa.pa_sample_spec:
        assert self._pa_stream is not None
        return pa.pa_stream_get_sample_spec(self._pa_stream)[0]

    def connect_playback(self, tlength: int = _UINT32_MAX, minreq: int = _UINT32_MAX) -> None:
        context = self.context()
        assert self._pa_stream is not None
        assert context is not None

        device = None

        buffer_attr = pa.pa_buffer_attr()
        buffer_attr.fragsize = _UINT32_MAX  # Irrelevant for playback
        buffer_attr.maxlength = _UINT32_MAX
        buffer_attr.tlength = tlength
        buffer_attr.prebuf = _UINT32_MAX
        buffer_attr.minreq = minreq

        flags = (pa.PA_STREAM_START_CORKED |
                 pa.PA_STREAM_INTERPOLATE_TIMING |
                 pa.PA_STREAM_VARIABLE_RATE)

        volume = None

        sync_stream = None  # TODO use this

        context.check(
            pa.pa_stream_connect_playback(self._pa_stream,
                                          device,
                                          buffer_attr,
                                          flags,
                                          volume,
                                          sync_stream)
        )

        while not self.is_ready and not self.is_failed:
            self.mainloop.wait()
        if not self.is_ready:
            context.raise_error()

        self._refresh_sink_index()
        # ba = pa.pa_stream_get_buffer_attr(self._pa_stream).contents
        # print(f"{ba.maxlength=}, {ba.tlength=}, {ba.prebuf=}, {ba.minreq=}, {ba.fragsize=}")

        assert _debug('PulseAudioStream: Playback connected')

    def begin_write(self, nbytes: Optional[int] = None) -> Tuple[ctypes.c_void_p, int]:
        context = self.context()
        assert context is not None

        addr = ctypes.c_void_p()
        nbytes_st = ctypes.c_size_t(_SIZE_T_MAX if nbytes is None else nbytes)

        context.check(
            pa.pa_stream_begin_write(self._pa_stream, ctypes.byref(addr), ctypes.byref(nbytes_st))
        )
        context.check_ptr_not_null(addr)

        assert _debug(f"PulseAudioStream: begin_write nbytes={nbytes} nbytes_n={nbytes_st.value}")

        return addr, nbytes_st.value

    def cancel_write(self) -> None:
        self.context().check(pa.pa_stream_cancel_write(self._pa_stream))

    def write(self, data, length: int, seek_mode=pa.PA_SEEK_RELATIVE) -> int:
        context = self.context()
        assert context is not None
        assert self._pa_stream is not None
        assert self.is_ready
        assert _debug(f'PulseAudioStream: writing {length} bytes')

        context.check(
            pa.pa_stream_write(self._pa_stream, data, length, pa.pa_free_cb_t(0), 0, seek_mode)
        )
        return length

    def update_timing_info(
        self,
        callback: Optional[PulseAudioContextSuccessCallback] = None,
    ) -> 'PulseAudioOperation':
        context = self.context()
        assert context is not None
        assert self._pa_stream is not None

        clump = PulseAudioStreamSuccessCallbackLump(context, callback)
        return PulseAudioOperation(
            clump,
            pa.pa_stream_update_timing_info(self._pa_stream, clump.pa_callback, None),
        )

    def get_timing_info(self) -> Optional[pa.pa_timing_info]:
        """
        Retrieves the stream's timing_info struct,
        or None if it does not exist.
        Note that ctypes creates a copy of the struct, meaning it will
        be safe to use with an unlocked mainloop.
        """
        context = self.context()
        assert context is not None
        assert self._pa_stream is not None
        timing_info = pa.pa_stream_get_timing_info(self._pa_stream)
        return timing_info.contents if timing_info else None

    def trigger(
        self,
        callback: Optional[PulseAudioContextSuccessCallback] = None,
    ) -> 'PulseAudioOperation':
        context = self.context()
        assert context is not None
        assert self._pa_stream is not None

        clump = PulseAudioStreamSuccessCallbackLump(context, callback)
        return PulseAudioOperation(
            clump,
            pa.pa_stream_trigger(self._pa_stream, clump.pa_callback, None),
        )

    def prebuf(
        self,
        callback: Optional[PulseAudioContextSuccessCallback] = None,
    ) -> 'PulseAudioOperation':
        context = self.context()
        assert context is not None
        assert self._pa_stream is not None

        clump = PulseAudioStreamSuccessCallbackLump(context, callback)
        return PulseAudioOperation(
            clump,
            pa.pa_stream_prebuf(self._pa_stream, clump.pa_callback, None),
        )

    def flush(
        self,
        callback: Optional[PulseAudioContextSuccessCallback] = None,
    ) -> 'PulseAudioOperation':
        context = self.context()
        assert context is not None
        assert self._pa_stream is not None

        clump = PulseAudioStreamSuccessCallbackLump(context, callback)
        return PulseAudioOperation(
            clump,
            pa.pa_stream_flush(self._pa_stream, clump.pa_callback, None),
        )

    def resume(
        self,
        callback: Optional[PulseAudioContextSuccessCallback] = None,
    ) -> 'PulseAudioOperation':
        return self._cork(False, callback)

    def pause(
        self,
        callback: Optional[PulseAudioContextSuccessCallback] = None,
    ) -> 'PulseAudioOperation':
        return self._cork(True, callback)

    def _cork(
        self,
        pause: Union[int, bool],
        callback: PulseAudioContextSuccessCallback,
    ) -> 'PulseAudioOperation':
        context = self.context()
        assert context is not None
        assert self._pa_stream is not None

        clump = PulseAudioStreamSuccessCallbackLump(context, callback)
        return PulseAudioOperation(
            clump,
            pa.pa_stream_cork(self._pa_stream, pause, clump.pa_callback, None),
        )

    def update_sample_rate(
        self,
        sample_rate: int,
        callback: Optional[PulseAudioContextSuccessCallback] = None,
    ) -> 'PulseAudioOperation':
        context = self.context()
        assert context is not None
        assert self._pa_stream is not None

        clump = PulseAudioStreamSuccessCallbackLump(context, callback)
        return PulseAudioOperation(
            clump,
            pa.pa_stream_update_sample_rate(
                self._pa_stream, sample_rate, clump.pa_callback, None
            ),
        )

    def set_write_callback(self, f: PulseAudioStreamRequestCallback) -> None:
        self._cb_write = pa.pa_stream_request_cb_t(f)
        pa.pa_stream_set_write_callback(self._pa_stream, self._cb_write, None)

    def set_underflow_callback(self, f: PulseAudioStreamNotifyCallback) -> None:
        self._cb_underflow = pa.pa_stream_notify_cb_t(f)
        pa.pa_stream_set_underflow_callback(self._pa_stream, self._cb_underflow, None)

    def _connect_callbacks(self) -> None:
        s = self._pa_stream
        pa.pa_stream_set_underflow_callback(s, self._cb_underflow, None)
        pa.pa_stream_set_write_callback(s, self._cb_write, None)
        pa.pa_stream_set_state_callback(s, self._cb_state, None)
        pa.pa_stream_set_moved_callback(s, self._cb_moved, None)

    def _disconnect_callbacks(self) -> None:
        s = self._pa_stream
        pa.pa_stream_set_underflow_callback(s, pa.pa_stream_notify_cb_t(0), None)
        pa.pa_stream_set_write_callback(s, pa.pa_stream_request_cb_t(0), None)
        pa.pa_stream_set_state_callback(s, pa.pa_stream_notify_cb_t(0), None)
        pa.pa_stream_set_moved_callback(s, pa.pa_stream_notify_cb_t(0), None)

    def _state_callback(self, _stream, _userdata) -> None:
        self._refresh_state()
        assert _debug(f"PulseAudioStream: state changed to {self._state_name[self.state]}")
        self.mainloop.signal()

    def _moved_callback(self, _stream, _userdata) -> None:
        self._refresh_sink_index()
        assert _debug(f"PulseAudioStream: moved to new index {self.index}")

    def _refresh_sink_index(self) -> None:
        self.index = pa.pa_stream_get_index(self._pa_stream)
        if self.index == PA_INVALID_INDEX:
            self.context().raise_error()

    def _refresh_state(self) -> None:
        self.state = pa.pa_stream_get_state(self._pa_stream)


class PulseAudioOperation(PulseAudioMainloopChild):
    """An asynchronous PulseAudio operation.
    Can be waited for, where it will run until completion or cancellation.
    Remember to `delete()` it with the mainloop lock held, otherwise
    it will be leaked.
    """

    _state_name = {pa.PA_OPERATION_RUNNING: 'Running',
                   pa.PA_OPERATION_DONE: 'Done',
                   pa.PA_OPERATION_CANCELLED: 'Cancelled'}

    def __init__(self, callback_lump, pa_operation: pa.pa_operation) -> None:
        context = callback_lump.context

        assert context.mainloop is not None
        assert pa_operation is not None
        context.check_ptr_not_null(pa_operation)

        super().__init__(context.mainloop)

        self.callback_lump = callback_lump
        self._pa_operation = pa_operation

    def _get_state(self) -> None:
        assert self._pa_operation is not None
        return pa.pa_operation_get_state(self._pa_operation)

    def delete(self) -> None:
        """Unref and delete the operation."""
        if self._pa_operation is not None:
            pa.pa_operation_unref(self._pa_operation)
            self._pa_operation = None
            self.callback_lump = None
            self.context = None

    def cancel(self):
        """Cancel the operation."""
        assert self._pa_operation is not None
        pa.pa_operation_cancel(self._pa_operation)
        return self

    def wait(self):
        """Wait until the operation is either done or cancelled."""
        while self.is_running:
            self.mainloop.wait()
        return self

    @property
    def is_running(self) -> bool:
        return self._get_state() == pa.PA_OPERATION_RUNNING

    @property
    def is_done(self) -> bool:
        return self._get_state() == pa.PA_OPERATION_DONE

    @property
    def is_cancelled(self) -> bool:
        return self._get_state() == pa.PA_OPERATION_CANCELLED


class PulseAudioContextSuccessCallbackLump:
    def __init__(
        self,
        context: PulseAudioContext,
        callback: Optional[PulseAudioContextSuccessCallback] = None,
    ) -> None:
        self.pa_callback = pa.pa_context_success_cb_t(self._success_callback)
        self.py_callback = callback
        self.context = context

    def _success_callback(self, context, success, userdata):
        if self.py_callback is not None:
            self.py_callback(context, success, userdata)
        self.context.mainloop.signal()


class PulseAudioStreamSuccessCallbackLump:
    def __init__(
        self,
        context: PulseAudioContext,
        callback: Optional[PulseAudioContextSuccessCallback] = None,
    ) -> None:
        self.pa_callback = pa.pa_stream_success_cb_t(self._success_callback)
        self.py_callback = callback
        self.context = context

    def _success_callback(self, stream, success, userdata):
        if self.py_callback is not None:
            self.py_callback(stream, success, userdata)
        self.context.mainloop.signal()



"""
Pythonic interface to DirectSound.
"""
import ctypes
import weakref
from collections import namedtuple

from pyglet.media.exceptions import MediaException
from pyglet.util import debug_print
from pyglet.window.win32 import _user32

from . import lib_dsound as lib
from .exceptions import DirectSoundNativeError

_debug = debug_print('debug_media')


def _check(hresult):
    if hresult != lib.DS_OK:
        raise DirectSoundNativeError(hresult)


def _create_wave_format(audio_format):
    if audio_format.channels > 2 or audio_format.sample_size not in (8, 16):
        raise MediaException(f'Unsupported audio format: {audio_format}')

    wfx = lib.WAVEFORMATEX()
    wfx.wFormatTag = lib.WAVE_FORMAT_PCM
    wfx.nChannels = audio_format.channels
    wfx.nSamplesPerSec = audio_format.sample_rate
    wfx.wBitsPerSample = audio_format.sample_size
    wfx.nBlockAlign = wfx.wBitsPerSample * wfx.nChannels // 8
    wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign
    return wfx


def _create_buffer_desc(wave_format, buffer_size):
    dsbdesc = lib.DSBUFFERDESC()
    dsbdesc.dwSize = ctypes.sizeof(dsbdesc)
    dsbdesc.dwFlags = (lib.DSBCAPS_GLOBALFOCUS |
                        lib.DSBCAPS_GETCURRENTPOSITION2 |
                        lib.DSBCAPS_CTRLFREQUENCY |
                        lib.DSBCAPS_CTRLVOLUME)
    if wave_format.nChannels == 1:
        dsbdesc.dwFlags |= lib.DSBCAPS_CTRL3D
    dsbdesc.dwBufferBytes = buffer_size
    dsbdesc.lpwfxFormat = ctypes.pointer(wave_format)

    return dsbdesc


def _create_primary_buffer_desc():
    """Primary buffer with 3D and volume capabilities"""
    buffer_desc = lib.DSBUFFERDESC()
    buffer_desc.dwSize = ctypes.sizeof(buffer_desc)
    buffer_desc.dwFlags = (lib.DSBCAPS_CTRL3D |
                            lib.DSBCAPS_CTRLVOLUME |
                            lib.DSBCAPS_PRIMARYBUFFER)

    return buffer_desc


class DirectSoundDriver:
    def __init__(self):
        assert _debug('Constructing DirectSoundDriver')

        self._native_dsound = lib.IDirectSound()
        _check(
            lib.DirectSoundCreate(None, ctypes.byref(self._native_dsound), None)
        )

        # A trick used by mplayer.. use desktop as window handle since it
        # would be complex to use pyglet window handles (and what to do when
        # application is audio only?).
        hwnd = _user32.GetDesktopWindow()
        _check(
            self._native_dsound.SetCooperativeLevel(hwnd, lib.DSSCL_NORMAL)
        )

        self.primary_buffer = self._create_primary_buffer()

    def delete(self):
        self.primary_buffer.delete()
        self.primary_buffer = None
        self._native_dsound.Release()

    def create_buffer(self, audio_format, buffer_size):
        wave_format = _create_wave_format(audio_format)
        buffer_desc = _create_buffer_desc(wave_format, buffer_size)
        return DirectSoundBuffer(
                self._create_native_buffer(buffer_desc),
                audio_format,
                buffer_size)

    def create_listener(self):
        return self.primary_buffer.create_listener()

    def _create_primary_buffer(self):
        return DirectSoundBuffer(
                self._create_native_buffer(_create_primary_buffer_desc()),
                None,
                0)

    def _create_native_buffer(self, buffer_desc):
        buf = lib.IDirectSoundBuffer()
        _check(
            self._native_dsound.CreateSoundBuffer(buffer_desc, ctypes.byref(buf), None)
        )
        return buf


_CurrentPosition = namedtuple('_CurrentPosition', ['play_cursor', 'write_cursor'])


class DirectSoundBuffer:
    def __init__(self, native_buffer, audio_format, buffer_size):
        self.audio_format = audio_format
        self.buffer_size = buffer_size

        self._native_buffer = native_buffer

        if audio_format is not None and audio_format.channels == 1:
            self._native_buffer3d = lib.IDirectSound3DBuffer()
            self._native_buffer.QueryInterface(lib.IID_IDirectSound3DBuffer,
                                               ctypes.byref(self._native_buffer3d))
        else:
            self._native_buffer3d = None

    def delete(self):
        if self._native_buffer is not None:
            self._native_buffer.Stop()
            self._native_buffer.Release()
            self._native_buffer = None
            if self._native_buffer3d is not None:
                self._native_buffer3d.Release()
                self._native_buffer3d = None

    @property
    def volume(self):
        vol = lib.LONG()
        _check(
            self._native_buffer.GetVolume(ctypes.byref(vol))
        )
        return vol.value

    @volume.setter
    def volume(self, value):
        _check(
            self._native_buffer.SetVolume(value)
        )

    @property
    def current_position(self):
        """Tuple of current play position and current write position.
        Only play position can be modified, so setter only accepts a single value."""
        play_cursor = lib.DWORD()
        write_cursor = lib.DWORD()
        _check(
            self._native_buffer.GetCurrentPosition(play_cursor,
                                                   write_cursor)
        )
        return _CurrentPosition(play_cursor.value, write_cursor.value)

    @current_position.setter
    def current_position(self, value):
        _check(
            self._native_buffer.SetCurrentPosition(value)
        )

    @property
    def is3d(self):
        return self._native_buffer3d is not None

    @property
    def is_playing(self):
        return (self._get_status() & lib.DSBSTATUS_PLAYING) != 0

    @property
    def is_buffer_lost(self):
        return (self._get_status() & lib.DSBSTATUS_BUFFERLOST) != 0

    def _get_status(self):
        status = lib.DWORD()
        _check(
            self._native_buffer.GetStatus(status)
        )
        return status.value

    @property
    def position(self):
        if self.is3d:
            position = lib.D3DVECTOR()
            _check(
                self._native_buffer3d.GetPosition(ctypes.byref(position))
            )
            return position.x, position.y, position.z
        else:
            return 0, 0, 0

    @position.setter
    def position(self, position):
        if self.is3d:
            x, y, z = position
            _check(
                self._native_buffer3d.SetPosition(x, y, z, lib.DS3D_IMMEDIATE)
            )

    @property
    def min_distance(self):
        """The minimum distance, which is the distance from the
        listener at which sounds in this buffer begin to be attenuated."""
        if self.is3d:
            value = lib.D3DVALUE()
            _check(
                self._native_buffer3d.GetMinDistance(ctypes.byref(value))
            )
            return value.value
        else:
            return 0

    @min_distance.setter
    def min_distance(self, value):
        if self.is3d:
            _check(
                self._native_buffer3d.SetMinDistance(value, lib.DS3D_IMMEDIATE)
            )

    @property
    def max_distance(self):
        """The maximum distance, which is the distance from the listener beyond which
        sounds in this buffer are no longer attenuated."""
        if self.is3d:
            value = lib.D3DVALUE()
            _check(
                self._native_buffer3d.GetMaxDistance(ctypes.byref(value))
            )
            return value.value
        else:
            return 0

    @max_distance.setter
    def max_distance(self, value):
        if self.is3d:
            _check(
                self._native_buffer3d.SetMaxDistance(value, lib.DS3D_IMMEDIATE)
            )

    @property
    def frequency(self):
        value = lib.DWORD()
        _check(
            self._native_buffer.GetFrequency(value)
        )
        return value.value

    @frequency.setter
    def frequency(self, value):
        """The frequency, in samples per second, at which the buffer is playing."""
        _check(
            self._native_buffer.SetFrequency(value)
        )

    @property
    def cone_orientation(self):
        """The orientation of the sound projection cone."""
        if self.is3d:
            orientation = lib.D3DVECTOR()
            _check(
                self._native_buffer3d.GetConeOrientation(ctypes.byref(orientation))
            )
            return orientation.x, orientation.y, orientation.z
        else:
            return 0, 0, 0

    @cone_orientation.setter
    def cone_orientation(self, value):
        if self.is3d:
            x, y, z = value
            _check(
                self._native_buffer3d.SetConeOrientation(x, y, z, lib.DS3D_IMMEDIATE)
            )

    _ConeAngles = namedtuple('_ConeAngles', ['inside', 'outside'])
    @property
    def cone_angles(self):
        """The inside and outside angles of the sound projection cone."""
        if self.is3d:
            inside = lib.DWORD()
            outside = lib.DWORD()
            _check(
                self._native_buffer3d.GetConeAngles(ctypes.byref(inside), ctypes.byref(outside))
            )
            return self._ConeAngles(inside.value, outside.value)
        else:
            return self._ConeAngles(0, 0)

    def set_cone_angles(self, inside, outside):
        """The inside and outside angles of the sound projection cone."""
        if self.is3d:
            _check(
                self._native_buffer3d.SetConeAngles(inside, outside, lib.DS3D_IMMEDIATE)
            )

    @property
    def cone_outside_volume(self):
        """The volume of the sound outside the outside angle of the sound projection cone."""
        if self.is3d:
            volume = lib.LONG()
            _check(
                self._native_buffer3d.GetConeOutsideVolume(ctypes.byref(volume))
            )
            return volume.value
        else:
            return 0

    @cone_outside_volume.setter
    def cone_outside_volume(self, value):
        if self.is3d:
            _check(
                self._native_buffer3d.SetConeOutsideVolume(value, lib.DS3D_IMMEDIATE)
            )

    def create_listener(self):
        native_listener = lib.IDirectSound3DListener()
        self._native_buffer.QueryInterface(lib.IID_IDirectSound3DListener,
                                           ctypes.byref(native_listener))
        return DirectSoundListener(native_listener)

    def play(self):
        _check(
            self._native_buffer.Play(0, 0, lib.DSBPLAY_LOOPING)
        )

    def stop(self):
        _check(
            self._native_buffer.Stop()
        )

    class _WritePointer:
        def __init__(self):
            self.audio_ptr_1 = ctypes.c_void_p()
            self.audio_length_1 = lib.DWORD()
            self.audio_ptr_2 = ctypes.c_void_p()
            self.audio_length_2 = lib.DWORD()

    def lock(self, write_cursor, write_size):
        assert _debug('DirectSoundBuffer.lock({}, {})'.format(write_cursor, write_size))
        pointer = self._WritePointer()
        _check(
            self._native_buffer.Lock(write_cursor,
                                     write_size,
                                     ctypes.byref(pointer.audio_ptr_1),
                                     ctypes.byref(pointer.audio_length_1),
                                     ctypes.byref(pointer.audio_ptr_2),
                                     ctypes.byref(pointer.audio_length_2),
                                     0)
        )
        return pointer

    def unlock(self, pointer):
        _check(
            self._native_buffer.Unlock(pointer.audio_ptr_1,
                                       pointer.audio_length_1,
                                       pointer.audio_ptr_2,
                                       pointer.audio_length_2)
        )


class DirectSoundListener:
    def __init__(self, native_listener):
        self._native_listener = native_listener

    def delete(self):
        if self._native_listener:
            self._native_listener.Release()
            self._native_listener = None

    @property
    def position(self):
        vector = lib.D3DVECTOR()
        _check(
            self._native_listener.GetPosition(ctypes.byref(vector))
        )
        return vector.x, vector.y, vector.z

    @position.setter
    def position(self, value):
        _check(
            self._native_listener.SetPosition(*(list(value) + [lib.DS3D_IMMEDIATE]))
        )

    @property
    def orientation(self):
        front = lib.D3DVECTOR()
        top = lib.D3DVECTOR()
        _check(
            self._native_listener.GetOrientation(ctypes.byref(front), ctypes.byref(top))
        )
        return front.x, front.y, front.z, top.x, top.y, top.z

    @orientation.setter
    def orientation(self, orientation):
        _check(
            self._native_listener.SetOrientation(*(list(orientation) + [lib.DS3D_IMMEDIATE]))
        )



"""Holds type aliases used throughout the codebase."""
import ctypes
import sys

from typing import Union, Literal


if sys.version_info >= (3, 12):
    from collections.abc import Buffer
else:
    # Best-effort placeholder for older Python versions
    Buffer = Union[bytes, bytearray, memoryview, ctypes.Array]

HorizontalAlign = Literal["left", "center", "right"]
AnchorX = Literal["left", "center", "right"]
AnchorY = Literal["top", "bottom", "center", "baseline"]
ContentVAlign = Literal["bottom", "center", "top"]



__all__ = [
    "Buffer", "HorizontalAlign", "AnchorX", "AnchorY", "ContentVAlign"
]



"""Information about version and extensions of current GLX implementation.

Usage::

    from pyglet.gl import glx_info

    if glx_info.have_extension('GLX_NV_float_buffer'):
        # ...

Or, if using more than one display::

    from pyglet.gl.glx_info import GLXInfo

    info = GLXInfo(window._display)
    if info.get_server_vendor() == 'ATI':
        # ...

"""

from __future__ import annotations

from ctypes import byref, c_int
from typing import TYPE_CHECKING

from pyglet.gl.glx import (
    GLX_EXTENSIONS,
    GLX_VENDOR,
    GLX_VERSION,
    glXGetClientString,
    glXQueryExtension,
    glXQueryExtensionsString,
    glXQueryServerString,
    glXQueryVersion,
)

from pyglet.util import asstr

if TYPE_CHECKING:
    from pyglet.libs.x11.xlib import Display


class GLXInfoException(Exception):  # noqa: D101, N818
    pass


class GLXInfo:  # noqa: D101
    def __init__(self, display: Display | None = None) -> None:  # noqa: D107
        # Set default display if not set
        if display and not _glx_info.display:
            _glx_info.set_display(display)

        self.display = display

    def set_display(self, display: Display) -> None:
        self.display = display

    def check_display(self) -> None:
        if not self.display:
            msg = 'No X11 display has been set yet.'
            raise GLXInfoException(msg)

    def have_version(self, major: int, minor: int = 0) -> bool:
        self.check_display()
        if not glXQueryExtension(self.display, None, None):
            msg = 'pyglet requires an X server with GLX'
            raise GLXInfoException(msg)

        server_version = self.get_server_version().split()[0]
        client_version = self.get_client_version().split()[0]

        server = [int(i) for i in server_version.split('.')]
        client = [int(i) for i in client_version.split('.')]
        return (tuple(server) >= (major, minor) and
                tuple(client) >= (major, minor))

    def get_server_vendor(self) -> str:
        self.check_display()
        return asstr(glXQueryServerString(self.display, 0, GLX_VENDOR))

    def get_server_version(self) -> str:
        # glXQueryServerString was introduced in GLX 1.1, so we need to use the
        # 1.0 function here which queries the server implementation for its
        # version.
        self.check_display()
        major = c_int()
        minor = c_int()
        if not glXQueryVersion(self.display, byref(major), byref(minor)):
            msg = 'Could not determine GLX server version'
            raise GLXInfoException(msg)
        return f'{major.value}.{minor.value}'

    def get_server_extensions(self) -> list[str]:
        self.check_display()
        return asstr(glXQueryServerString(self.display, 0, GLX_EXTENSIONS)).split()

    def get_client_vendor(self) -> str:
        self.check_display()
        return asstr(glXGetClientString(self.display, GLX_VENDOR))

    def get_client_version(self) -> str:
        self.check_display()
        return asstr(glXGetClientString(self.display, GLX_VERSION))

    def get_client_extensions(self) -> list[str]:
        self.check_display()
        return asstr(glXGetClientString(self.display, GLX_EXTENSIONS)).split()

    def get_extensions(self) -> list[str]:
        self.check_display()
        return asstr(glXQueryExtensionsString(self.display, 0)).split()

    def have_extension(self, extension: str) -> bool:
        self.check_display()
        if not self.have_version(1, 1):
            return False
        return extension in self.get_extensions()


# Single instance suitable for apps that use only a single display.
_glx_info = GLXInfo()

set_display = _glx_info.set_display
check_display = _glx_info.check_display
have_version = _glx_info.have_version
get_server_vendor = _glx_info.get_server_vendor
get_server_version = _glx_info.get_server_version
get_server_extensions = _glx_info.get_server_extensions
get_client_vendor = _glx_info.get_client_vendor
get_client_version = _glx_info.get_client_version
get_client_extensions = _glx_info.get_client_extensions
get_extensions = _glx_info.get_extensions
have_extension = _glx_info.have_extension



"""Load application resources from a known path.

Loading resources by specifying relative paths to filenames is often
problematic in Python, as the working directory is not necessarily the same
directory as the application's script files.

This module allows applications to specify a search path for resources.
Relative paths are taken to be relative to the application's ``__main__``
module. ZIP files can appear on the path; they will be searched inside.  The
resource module also behaves as expected when applications are bundled using
Freezers such as PyInstaller, Nuitka, py2app, etc..

In addition to providing file references (with the :py:func:`file` function),
the resource module also contains convenience functions for loading images,
textures, fonts, media and documents.

3rd party modules or packages not bound to a specific application should
construct their own :py:class:`Loader` instance and override the path to use the
resources in the module's directory.

Path format
^^^^^^^^^^^

The resource path :py:attr:`path` (see also :py:meth:`Loader.__init__` and
:py:meth:`Loader.path`)
is a list of locations to search for resources.  Locations are searched in the
order given in the path.  If a location is not valid (for example, if the
directory does not exist), it is skipped.

Locations in the path beginning with an "at" symbol (''@'') specify
Python packages.  Other locations specify a ZIP archive or directory on the
filesystem.  Locations that are not absolute are assumed to be relative to the
script home.  Some examples::

    # Search just the `res` directory, assumed to be located alongside the
    # main script file.
    path = ['res']

    # Search the directory containing the module `levels.level1`, followed
    # by the `res/images` directory.
    path = ['@levels.level1', 'res/images']

Paths are always **case-sensitive** and **forward slashes are always used**
as path separators, even in cases when the filesystem or platform does not do this.
This avoids a common programmer error when porting applications between platforms.

The default path is ``['.']``.  If you modify the path, you must call
:py:func:`reindex`.
"""
from __future__ import annotations

import os
import sys
import zipfile
import weakref

from io import BytesIO, StringIO
from typing import TYPE_CHECKING, IO

import pyglet

if TYPE_CHECKING:
    from typing import Literal
    from pyglet.graphics import Batch
    from pyglet.graphics.shader import Shader
    from pyglet.image import AbstractImage, Texture, TextureRegion
    from pyglet.image.animation import Animation
    from pyglet.image.atlas import TextureBin
    from pyglet.media.codecs import Source
    from pyglet.model import Scene
    from pyglet.text.document import AbstractDocument


class ResourceNotFoundException(Exception):
    """The named resource was not found on the search path."""

    def __init__(self, name):
        message = (f"Resource '{name}' was not found on the path.  "
                   "Ensure that the filename has the correct capitalisation.")
        Exception.__init__(self, message)


class UndetectableShaderType(Exception):
    """The type of the Shader source could not be identified."""

    def __init__(self, name):
        message = (f"The Shader type of '{name}' could not be determined. "
                   "Ensure that your source file has a standard extension, "
                   "or provide a valid 'shader_type' parameter.")
        Exception.__init__(self, message)


def get_script_home() -> str:
    """Get the directory containing the program entry module.

    For ordinary Python scripts, this is the directory containing the
    ``__main__`` module. For applications that have been bundled with
    PyInstaller, Nuitka, etc., this may be the bundle path or a
    temporary directory.

    If none of the above cases apply and the file for ``__main__`` cannot
    be determined the working directory is returned.

    When the script is being run by a Python profiler, this function
    may return the directory where the profiler is running instead of
    the directory of the real script. To work around this behaviour the
    full path to the real script can be specified in :py:attr:`pyglet.resource.path`.
    """
    frozen = getattr(sys, 'frozen', None)
    meipass = getattr(sys, '_MEIPASS', None)
    if meipass:
        # PyInstaller
        return meipass
    elif frozen in ('windows_exe', 'console_exe'):
        return os.path.dirname(sys.executable)
    elif frozen == 'macosx_app':
        # py2app
        return os.environ['RESOURCEPATH']
    else:
        main = sys.modules['__main__']
        if hasattr(main, '__file__'):
            return os.path.dirname(os.path.abspath(main.__file__))
        else:
            if 'python' in os.path.basename(sys.executable):
                # interactive
                return os.getcwd()
            else:
                # cx_Freeze
                return os.path.dirname(sys.executable)


def get_settings_path(name: str) -> str:
    """Get a directory path to save user preferences.

    Different platforms have different conventions for where to save user
    preferences and settings. This function implements those conventions
    as described below, and returns a fully formed path.

    On Linux, a directory ``name`` in the user's configuration directory is
    returned (usually under ``~/.config``).

    On Windows (including under Cygwin) the ``name`` directory in the user's
    ``Application Settings`` directory is returned.

    On Mac OS X the ``name`` directory under ``~/Library/Application Support``
    is returned.

    .. note:: This function does not perform any directory creation. Users
              should use ``os.path.exists`` and ``os.makedirs`` to construct
              the directory if desired.
    """
    if pyglet.compat_platform in ('cygwin', 'win32'):
        if 'APPDATA' in os.environ:
            return os.path.join(os.environ['APPDATA'], name)
        else:
            return os.path.expanduser(f'~/{name}')
    elif pyglet.compat_platform == 'darwin':
        return os.path.expanduser(f'~/Library/Application Support/{name}')
    elif pyglet.compat_platform.startswith('linux'):
        if 'XDG_CONFIG_HOME' in os.environ:
            return os.path.join(os.environ['XDG_CONFIG_HOME'], name)
        else:
            return os.path.expanduser(f'~/.config/{name}')
    else:
        return os.path.expanduser(f'~/.{name}')


def get_data_path(name: str) -> str:
    """Get a directory to save user data.

    For a Posix or Linux based system many distributions have a separate
    directory to store user data for a specific application and this 
    function returns the path to that location.

    On Linux, a directory ``name`` in the user's data directory is returned
    (usually under ``~/.local/share``).

    On Windows (including under Cygwin) the ``name`` directory in the user's
    ``Application Settings`` directory is returned.

    On Mac OS X the ``name`` directory under ``~/Library/Application Support``
    is returned.

    .. note:: This function does not perform any directory creation. Users
              should use ``os.path.exists`` and ``os.makedirs`` to construct
              the directory if desired.
    """
    if pyglet.compat_platform in ('cygwin', 'win32'):
        if 'APPDATA' in os.environ:
            return os.path.join(os.environ['APPDATA'], name)
        else:
            return os.path.expanduser(f'~/{name}')
    elif pyglet.compat_platform == 'darwin':
        return os.path.expanduser(f'~/Library/Application Support/{name}')
    elif pyglet.compat_platform.startswith('linux'):
        if 'XDG_DATA_HOME' in os.environ:
            return os.path.join(os.environ['XDG_DATA_HOME'], name)
        else:
            return os.path.expanduser(f'~/.local/share/{name}')
    else:
        return os.path.expanduser(f'~/.{name}')


class Location:
    """Abstract resource location.

    Given a location, a file can be loaded from that location with the
    :py:meth:`open` method. This provides a convenient way to specify a
    path to load files from, even when that path does not reside on the
    filesystem.
    """

    def open(self, name: str, mode: str = 'rb') -> BytesIO | StringIO | IO:
        """Open a file at this location.

        Args:
            name:
                The file name to open. Absolute paths are not supported.
                Relative paths are not supported by most locations (you
                should specify only a file name with no path component).
            mode:
                The file mode to open with.  Only files opened on the
                filesystem make use of this parameter; others ignore it.
        """
        raise NotImplementedError('abstract')


class FileLocation(Location):
    """Location on the filesystem."""

    def __init__(self, filepath: str) -> None:
        """Create a location given a relative or absolute path."""
        self.path = filepath

    def open(self, filename: str, mode: str = 'rb') -> IO:
        return open(os.path.join(self.path, filename), mode)


class ZIPLocation(Location):
    """Location within a ZIP file."""

    def __init__(self, zipfileobj: zipfile.ZipFile, directory: str | None):
        """Create a location given an open ZIP file and a path within that file.

        Args:
            zipfileobj:
                An open ZIP file from the ``zipfile`` module.
            directory:
                A path within that ZIP file.  Can be empty to specify files at
                the top level of the ZIP file.
        """
        self.zip = zipfileobj
        self.dir = directory

    def open(self, filename: str, mode='rb') -> BytesIO | StringIO:
        """Open a file from inside the ZipFile.

        Args:
            filename:
                The filename to open.
            mode:
                Valid modes are 'r' and 'rb'.
        """
        _path = f"{self.dir}/{filename}" if self.dir else filename
        _forward_slash_path = _path.replace(os.sep, '/')  # zip can only handle forward slashes
        _bytes = self.zip.read(_forward_slash_path)
        if mode == 'r':
            return StringIO(_bytes.decode())
        return BytesIO(_bytes)

    def __repr__(self):
        return f"{self.__class__.__name__}(dir='{self.dir}')"


class URLLocation(Location):
    """Location on the network.

    This class uses the ``urllib`` module to open files on
    the network, given a base URL.
    """

    def __init__(self, base_url: str) -> None:
        """Create a location given a base URL."""
        self.base = base_url

    def open(self, filename: str, mode: str = '') -> IO:
        """Open a remote file.

        Args:
            filename:
                The name of the remote resource to open.
            mode:
                Unused, as the mode is determined by the remote server.
        """
        import urllib.parse
        import urllib.request
        url = urllib.parse.urljoin(self.base, filename)
        return BytesIO(urllib.request.urlopen(url).read())


class Loader:
    """Load program resource files from disk.

    The loader contains a search path which can include filesystem
    directories, ZIP archives, URLs, and Python packages.
    """
    def __init__(self, pathlist: list[str] | None = None, script_home: str | None = None) -> None:
        """Create a loader for the given path.

        If no path is specified it defaults to ``['.']``; that is,
        just the program directory.

        See the module documentation for details on the path format.

        Args:
            pathlist:
                List of locations to search for resources.
            script_home:
                Base location of relative files. Defaults to
                the result of :py:func:`get_script_home`.
        """
        pathlist = pathlist or ['.']

        if isinstance(pathlist, str):
            pathlist = [pathlist]

        self.path = list(pathlist)
        self._script_home = script_home or get_script_home()
        self._index: dict | None = None

        # Map bin size to list of atlases
        self._texture_atlas_bins = {}

        # map name to image etc.
        self._cached_textures = weakref.WeakValueDictionary()
        self._cached_images = weakref.WeakValueDictionary()
        self._cached_animations = weakref.WeakValueDictionary()

    def _ensure_index(self):
        if self._index is None:
            self.reindex()

    def _index_file(self, name: str, locationobj: Location) -> None:
        if name not in self._index:
            self._index[name] = locationobj

    def reindex(self):
        """Refresh the file index.

        You must call this method if ``resource.path`` is changed,
        or the filesystem layout changes.
        """
        self._index = {}
        for _path_name in self.path:

            # A Python module:
            if _path_name.startswith('@'):
                module_name = _path_name[1:]
                try:
                    module = __import__(module_name)
                except (ImportError, ValueError):
                    continue
                for component in module_name.split('.')[1:]:
                    module = getattr(module, component)
                if hasattr(module, '__file__'):
                    _path_name = os.path.dirname(module.__file__)
                else:
                    _path_name = ''  # interactive

            elif not os.path.isabs(_path_name):
                # Add script base unless absolute
                assert r'\\' not in _path_name, "Backslashes are not permitted in relative paths"
                _path_name = os.path.join(self._script_home, _path_name)

            # A filesystem directory:
            if os.path.isdir(_path_name):
                _path_name = _path_name.rstrip(os.path.sep)
                # os.walk will pass silently if the path is not found:
                for dirpath, dirnames, filenames in os.walk(_path_name):
                    dirpath = dirpath[len(_path_name) + 1:]
                    # Force forward slashes for index
                    if dirpath:
                        parts = [part for part in dirpath.split(os.sep) if part is not None]
                        dirpath = '/'.join(parts)
                    for filename in filenames:
                        if dirpath:
                            index_name = dirpath + '/' + filename
                        else:
                            index_name = filename
                        file_location = FileLocation(_path_name)
                        self._index_file(index_name, file_location)

            else:
                # The path is not a valid directory. See if it's a
                # ZIP file, or a nested ZIP file with internal paths.
                # Ex:  path/to/file.zip
                #      path/to/file.zip/internal/path
                zip_directory = ''
                old_path = None
                while _path_name and not (os.path.isfile(_path_name)):
                    old_path = _path_name
                    _path_name, tail_dir = os.path.split(_path_name)
                    if _path_name == old_path:
                        break
                    zip_directory = '/'.join((tail_dir, zip_directory))
                if _path_name == old_path:
                    continue

                zip_directory = zip_directory.rstrip('/')

                if zipfile.is_zipfile(_path_name):
                    zipfileobj = zipfile.ZipFile(_path_name, 'r')
                    # Returns zipfile.ZipInfo objects:
                    for fileinfo in zipfileobj.infolist():
                        if fileinfo.is_dir():
                            continue
                        if zip_directory not in fileinfo.filename:
                            continue

                        filename = fileinfo.filename.lstrip(zip_directory)
                        filename = filename.lstrip('/')

                        file_location = ZIPLocation(zipfileobj, zip_directory)
                        self._index_file(filename, file_location)

    def file(self, name: str, mode: str = 'rb') -> BytesIO | StringIO | IO:
        """Load a file-like object.

        Args:
            name:
                Filename of the resource to load.
            mode:
                Combination of ``r``, ``w``, ``a``, ``b`` and ``t`` characters
                with the meaning as for the builtin ``open`` function.
        """
        self._ensure_index()
        try:
            file_location = self._index[name]
            return file_location.open(name, mode)
        except KeyError:
            raise ResourceNotFoundException(name)

    def location(self, filename: str) -> FileLocation | URLLocation | ZIPLocation:
        """Get the location of a resource.

        This method is useful for opening files referenced from a resource.
        For example, an HTML file loaded as a resource might reference some
        images.  These images should be located relative to the HTML file, not
        looked up individually in the loader's path.
        """
        self._ensure_index()
        try:
            return self._index[filename]
        except KeyError:
            raise ResourceNotFoundException(filename)

    def add_font(self, filename: str) -> None:
        """Add a font resource to the application.

        Fonts not installed on the system must be added to pyglet before they
        can be used with ``font.load``. Although the font is added with its
        filename using this function, fonts are always loaded by specifying
        their family name. For example::

            resource.add_font('action_man.ttf')
            action_man = font.load('Action Man')

        """
        self._ensure_index()
        from pyglet import font
        fileobj = self.file(filename)
        font.add_file(fileobj)

    def _alloc_image(self, name: str, use_atlas: bool, border: int) -> AbstractImage:
        fileobj = self.file(name)
        try:
            img = pyglet.image.load(name, file=fileobj)
        finally:
            fileobj.close()

        if not use_atlas:
            return img.get_texture()

        # Add the image to a TextureAtlasBin, if possible
        if texture_bin := self._get_texture_atlas_bin(img.width, img.height, border):
            return texture_bin.add(img, border)

        return img.get_texture()

    def _get_texture_atlas_bin(self, width: int, height: int, border: int) -> TextureBin | None:
        """A heuristic for determining the atlas bin to use for a given image
        size.  Returns None if the image should not be placed in an atlas (too
        big), otherwise the bin (a list of TextureAtlas).
        """
        # Large images are not placed in an atlas
        max_texture_size = pyglet.image.get_max_texture_size()
        max_size = min(2048, max_texture_size) - border
        if width > max_size or height > max_size:
            return None

        # Group images with small height separately to larger height
        # (as the allocator can't stack within a single row).
        bin_size = 1
        if height > max_size / 4:
            bin_size = 2

        try:
            texture_bin = self._texture_atlas_bins[bin_size]
        except KeyError:
            texture_bin = pyglet.image.atlas.TextureBin()
            self._texture_atlas_bins[bin_size] = texture_bin

        return texture_bin

    def image(self, name: str, flip_x: bool = False, flip_y: bool = False, rotate: Literal[0, 90, 180, 270, 360] = 0,
              atlas: bool = True, border: int = 1) -> Texture | TextureRegion:
        """Load an image with optional transformation.

        This is similar to `texture`, except the resulting image will be
        packed into a :py:class:`~pyglet.image.atlas.TextureBin` (TextureAtlas)
        if it is an appropriate size for packing. This is more efficient than
        loading images into separate textures.

        Args:
            name:
                The filename of the image source to load.
            flip_x:
                If ``True``, the returned image will be flipped horizontally.
            flip_y:
                If ``True``, the returned image will be flipped vertically.
            rotate:
                The returned image will be rotated clockwise by the given
                number of degrees (a multiple of 90).
            atlas:
                If ``True``, the image will be loaded into an atlas managed by
                pyglet. If atlas loading is not appropriate for specific texturing
                reasons (e.g. border control is required) then set to ``False``.
            border:
                Leaves specified pixels of blank space around each image in
                an atlas, which may help reduce texture bleeding.

        .. note:: When using ``flip_x/y`` or ``rotate``, the actual image
                  data is not modified. Instead, the texture coordinates
                  are manipulated to produce the desired result.
        """
        self._ensure_index()
        if name in self._cached_images:
            identity = self._cached_images[name]
        else:
            identity = self._cached_images[name] = self._alloc_image(name, atlas, border)

        if not rotate and not flip_x and not flip_y:
            return identity

        return identity.get_transform(flip_x, flip_y, rotate)

    def animation(self, name: str, flip_x: bool = False, flip_y: bool = False,
                  rotate: Literal[0, 90, 180, 270, 360] = 0, border: int = 1) -> Animation:
        """Load an animation with optional transformation.

        Animations loaded from the same source but with different
        transformations will use the same textures.

        Args:
            name:
                Filename of the animation source to load.
            flip_x:
                If ``True``, the returned image will be flipped horizontally.
            flip_y:
                If ``True``, the returned image will be flipped vertically.
            rotate:
                The returned image will be rotated clockwise by the given
                number of degrees (must be a multiple of 90).
            border:
                Leaves specified pixels of blank space around each image in
                an atlas, which may help reduce texture bleeding.
        """
        self._ensure_index()
        try:
            identity = self._cached_animations[name]
        except KeyError:
            _animation = pyglet.image.load_animation(name, self.file(name))
            texture_bin = self._get_texture_atlas_bin(_animation.get_max_width(),
                                                      _animation.get_max_height(),
                                                      border)
            if texture_bin:
                _animation.add_to_texture_bin(texture_bin, border)

            identity = self._cached_animations[name] = _animation

        if not rotate and not flip_x and not flip_y:
            return identity

        return identity.get_transform(flip_x, flip_y, rotate)

    def get_cached_animation_names(self) -> list[str]:
        """Get a list of animation filenames that have been cached.

        This is useful for debugging and profiling only.
        """
        self._ensure_index()
        return list(self._cached_animations.keys())

    def get_cached_image_names(self) -> list[str]:
        """Get a list of image filenames that have been cached.

        This is useful for debugging and profiling only.
        """
        self._ensure_index()
        return list(self._cached_images.keys())

    def get_cached_texture_names(self) -> list[str]:
        """Get a list of texture filenames that have been cached.

        This is useful for debugging and profiling only.
        """
        self._ensure_index()
        return list(self._cached_textures.keys())

    def get_texture_bins(self) -> list[TextureBin]:
        """Get a list of texture bins in use.

        This is useful for debugging and profiling only.
        """
        self._ensure_index()
        return list(self._texture_atlas_bins.values())

    def media(self, name: str, streaming: bool = True) -> Source:
        """Load a sound or video resource.

        The meaning of ``streaming`` is as for :py:func:`~pyglet.media.load`.
        Compressed sources cannot be streamed (that is, video and compressed
        audio cannot be streamed from a ZIP archive).

        Args:
            name:
                Filename of the media source to load.
            streaming:
                True if the source should be streamed from disk, False if
                it should be entirely decoded into memory immediately.
        """
        self._ensure_index()
        from pyglet import media
        try:
            file_location = self._index[name]
            if isinstance(location, FileLocation):
                # Don't open the file if it's streamed from disk
                file_path = os.path.join(file_location.path, name)
                return media.load(file_path, streaming=streaming)
            else:
                fileobj = file_location.open(name)

                return media.load(name, file=fileobj, streaming=streaming)
        except KeyError:
            raise ResourceNotFoundException(name)

    def texture(self, name: str) -> Texture:
        """Load an image as a single OpenGL texture."""
        self._ensure_index()
        if name in self._cached_textures:
            return self._cached_textures[name]

        fileobj = self.file(name)
        textureobj = pyglet.image.load(name, file=fileobj).get_texture()
        self._cached_textures[name] = textureobj
        return textureobj

    def scene(self, name: str) -> Scene:
        """Load a 3D Scene."""
        self._ensure_index()
        abspathname = os.path.join(os.path.abspath(self.location(name).path), name)
        return pyglet.model.load(filename=abspathname, file=self.file(name))

    def html(self, name: str) -> AbstractDocument:
        """Load an HTML document."""
        self._ensure_index()
        fileobj = self.file(name)
        return pyglet.text.load(name, fileobj, 'text/html')

    def attributed(self, name: str) -> AbstractDocument:
        """Load an attributed text document.

        See `pyglet.text.formats.attributed` for details on this format.
        """
        self._ensure_index()
        fileobj = self.file(name)
        return pyglet.text.load(name, fileobj, 'text/vnd.pyglet-attributed')

    def text(self, name: str) -> AbstractDocument:
        """Load a plain text document."""
        self._ensure_index()
        fileobj = self.file(name)
        return pyglet.text.load(name, fileobj, 'text/plain')

    def shader(self, name: str, shader_type: str | None = None) -> Shader:
        """Load a Shader object.

        Args:
            name:
                Filename of the Shader source to load.
            shader_type:
                A hint for the type of shader, such as 'vertex', 'fragment', etc.
                Not required if your shader has a standard file extension, such
                as ``.vert``, ``.frag``, etc..
        """
        self._ensure_index()
        # https://www.khronos.org/opengles/sdk/tools/Reference-Compiler/
        shader_extensions = {'comp': "compute",
                             'frag': "fragment",
                             'geom': "geometry",
                             'tesc': "tescontrol",
                             'tese': "tesevaluation",
                             'vert': "vertex"}
        fileobj = self.file(name, 'r')
        source_string = fileobj.read()

        if not shader_type:
            try:
                _, extension = os.path.splitext(name)
                shader_type = shader_extensions[extension.strip(".")]
            except KeyError:
                raise UndetectableShaderType(name=name)

        if shader_type not in shader_extensions.values():
            raise UndetectableShaderType(name=name)

        return pyglet.graphics.shader.Shader(source_string, shader_type)


#: Default resource search path.
#:
#: Locations in the search path are searched in order and are always
#: case-sensitive.  After changing the path you must call `reindex`.
#:
#: See the module documentation for details on the path format.
#:
#: :type: list of str
path = []


class _DefaultLoader(Loader):

    @property
    def path(self):
        return path

    @path.setter
    def path(self, value):
        global path
        path = value


_default_loader = _DefaultLoader()
reindex = _default_loader.reindex
file = _default_loader.file
location = _default_loader.location
add_font = _default_loader.add_font
image = _default_loader.image
animation = _default_loader.animation
media = _default_loader.media
texture = _default_loader.texture
html = _default_loader.html
attributed = _default_loader.attributed
text = _default_loader.text
scene = _default_loader.scene
shader = _default_loader.shader
get_cached_texture_names = _default_loader.get_cached_texture_names
get_cached_image_names = _default_loader.get_cached_image_names
get_cached_animation_names = _default_loader.get_cached_animation_names
get_texture_bins = _default_loader.get_texture_bins



"""FFmpeg wrapping"""
from .libavcodec import *
from .libavutil import *
from .libavformat import *
from .libswresample import *
from .libswscale import *
from .compat import *
apply_version_changes()



"""
Using PyGLM/glm to verify commonly supported operators and functions.
"""
import math

import pytest
from pyglet.math import Vec2

# mix
# min / max
# angle (between two vectors)
# from_heading (from angle to vector)


def test_create():
    assert Vec2
    assert Vec2(1, 2) == Vec2(1, 2)
    # assert Vec2(1) == Vec2(1, 1)  # not supported
    with pytest.raises(TypeError):
        Vec2(1, 2, 3)


def test_access():
    """Test Vec2 class."""
    # Basic access
    v = Vec2(1, 2)
    assert v.x == 1
    assert v.y == 2
    assert v[0] == 1
    assert v[1] == 2
    assert v == (1, 2)


def test_comparison():
    assert Vec2(1, 2) == Vec2(1, 2)
    assert Vec2(1, 2) != Vec2(2, 1)
    assert Vec2(1, 2) == (1, 2)
    assert (1, 2) == Vec2(1, 2)


def test_swizzle():
    """Test swizzle."""
    v = Vec2(1, 2)
    assert v.xy == (1, 2)
    assert v.yx == (2, 1)
    assert v.xx == (1, 1)
    assert v.yy == (2, 2)
    assert v.xyxy == (1, 2, 1, 2)

    with pytest.raises(AttributeError):
        v.xxxxx


def test_mutability():
    v = Vec2(1, 2)
    with pytest.raises(AttributeError):
        v.x = 1
    with pytest.raises(AttributeError):
        v.y = 1
    with pytest.raises(TypeError):
        v[0] = 1  # __setitem__ is not supported

    # Swizzle is an output-only operation
    with pytest.raises(AttributeError):
        v.xy = (1, 2)
    with pytest.raises(AttributeError):
        v.yx = (2, 1)
    with pytest.raises(AttributeError):
        v.xx = (1, 1)
    with pytest.raises(AttributeError):
        v.yy = (2, 2)
    with pytest.raises(AttributeError):
        v.xyxy = (1, 2, 1, 2)


def test_len():
    """Len of the collection is always 2."""
    assert len(Vec2(0)) == 2


def test_max_min():
    assert max(Vec2(1, 2)) == 2
    assert max(Vec2(1, -2)) == 1

    assert min(Vec2(1, 2)) == 1
    assert min(Vec2(1, -2)) == -2


def test_add():
    assert Vec2(1, 2) + Vec2(3, 4) == Vec2(4, 6)

    assert Vec2(1, 2) + 3 == Vec2(4, 5)
    assert 2 + Vec2(1, 2) == Vec2(3, 4)

    assert Vec2(1, 2) + (3, 4) == Vec2(4, 6)
    assert (1, 2) + Vec2(3, 4) == Vec2(4, 6)

    v = Vec2(1, 2)
    v += 1
    assert v == Vec2(2, 3)

    v = Vec2(1, 3)
    v += Vec2(3, 4)
    assert v == Vec2(4, 7)

    v = Vec2(1, 2)
    v += (3, 4)
    assert v == Vec2(4, 6)


def test_radd():
    # These should be distinct and  unreachable through
    # simple mis-addition of the mirrored values below.
    goal_x = -1901.0
    goal_y = 7919.0
    pairs = [
        # A few (+even, -odd) Vec2s + tuples
        Vec2( -8.0,     7.0),
        ( -6.0,     5.0),
        Vec2( -4.0,     3.0),
        ( -2.0,     1.0),

        Vec2(goal_x, goal_y),

        # Negatives of the block above the marker line
        (  2.0, -    1.0),
        Vec2(  4.0, -    3.0),
        (  6.0, -    5.0),
        Vec2(  8.0, -    7.0),
    ]
    assert sum(pairs) == Vec2(goal_x, goal_y)


def test_sub():
    assert Vec2(1, 2) - Vec2(3, 4) == Vec2(-2, -2)

    assert Vec2(1, 2) - 3 == Vec2(-2, -1)
    assert 2 - Vec2(1, 2) == Vec2(1, 0)

    assert Vec2(1, 2) - (3, 4) == Vec2(-2, -2)
    assert (1, 2) - Vec2(3, 4) == Vec2(-2, -2)

    v = Vec2(1, 2)
    v -= Vec2(1, 1)
    assert v == Vec2(0, 1)

    v = Vec2(1, 2)
    v -= (1, 2)
    assert v == Vec2(0, 0)

    v = Vec2(0, 1)
    v -= 1
    assert v == Vec2(-1, 0)


def test_mul():
    assert Vec2(2, 3) * Vec2(4, 5) == Vec2(8, 15)

    assert Vec2(1, 2) * 2 == Vec2(2, 4)
    assert 2 * Vec2(1, 2) == Vec2(2, 4)

    assert Vec2(1, 2) * (2, 2) == Vec2(2, 4)
    assert (1, 2) * Vec2(2, 2) == Vec2(2, 4)

    v = Vec2(1, 2)
    v *= Vec2(2, 2)
    assert v == Vec2(2, 4)

    v = Vec2(1, 2)
    v *= 2
    assert v == Vec2(2, 4)

    v = Vec2(1, 2)
    v *= Vec2(2, 2)
    assert v == Vec2(2, 4)


def test_truediv():
    assert Vec2(2, 4) / Vec2(2, 2) == Vec2(1, 2)

    assert Vec2(2, 4) / 2.0 == Vec2(1, 2)
    assert 1.0 / Vec2(0.5, 0.5) == Vec2(2, 2)

    assert (2, 4) / Vec2(2, 2) == Vec2(1, 2)
    assert Vec2(2, 4) / (2, 2) == Vec2(1, 2)

    v = Vec2(2, 4)
    v /= Vec2(2, 2)
    assert v == Vec2(1, 2)

    v = Vec2(2, 4)
    v /= 2
    assert v == Vec2(1, 2)

    v = Vec2(2, 4)
    v /= (2, 2)
    assert v == Vec2(1, 2)

    with pytest.raises(ZeroDivisionError):
        Vec2(1, 2) / Vec2(0, 0)
    with pytest.raises(ZeroDivisionError):
        Vec2(1, 2) / (0, 0)
    with pytest.raises(ZeroDivisionError):
        (1, 2) / Vec2(0, 0)
    with pytest.raises(ZeroDivisionError):
        Vec2(1, 2) / 0
    with pytest.raises(ZeroDivisionError):
        v = Vec2(1, 2)
        v /= 0


def test_floordiv():
    assert Vec2(2.5, 4.5) // Vec2(2, 2) == Vec2(1, 2)

    assert Vec2(2.5, 4.5) // 2 == Vec2(1, 2)
    assert 4.2 // Vec2(2, 4) == Vec2(2, 1)

    assert (1, 2) // Vec2(0.5, 0.5) == Vec2(2, 4)
    assert Vec2(1, 2) // (0.5, 0.5) == Vec2(2, 4)

    v = Vec2(4, 8)
    v //= 2
    assert v == Vec2(2, 4)

    v = Vec2(4, 8)
    v //= Vec2(2, 2)
    assert v == Vec2(2, 4)

    v = Vec2(4, 8)
    v //= (2, 2)
    assert v == Vec2(2, 4)

    with pytest.raises(ZeroDivisionError):
        Vec2(1, 2) // Vec2(0, 0)
    with pytest.raises(ZeroDivisionError):
        Vec2(1, 2) // (0, 0)
    with pytest.raises(ZeroDivisionError):
        (1, 2) // Vec2(0, 0)
    with pytest.raises(ZeroDivisionError):
        Vec2(1, 2) // 0
    with pytest.raises(ZeroDivisionError):
        v = Vec2(1, 2)
        v //= 0


def test_length():
    assert Vec2(1, 0).length() == 1
    assert Vec2(0, 1).length() == 1

    assert Vec2(3, 0).length() == 3
    assert Vec2(0, 3).length() == 3

    assert Vec2(1, 1).length() == pytest.approx(1.41421, abs=1e-5)
    assert Vec2(-1, -1).length() == pytest.approx(1.41421, abs=1e-5)


def test_length_squared():
    assert Vec2(1, 0).length_squared() == 1
    assert Vec2(0, -1).length_squared() == 1

    assert Vec2(3, 0).length_squared() == 9
    assert Vec2(0, 3).length_squared() == 9

    assert Vec2(1, 1).length_squared() == 2
    assert Vec2(-1, -1).length_squared() == 2


def test_abs():
    assert abs(Vec2(1, 2)) == Vec2(1, 2)
    assert abs(Vec2(-1, -2)) == Vec2(1, 2)


def test_neg():
    """Negate the vector."""
    assert -Vec2(1, 2) == Vec2(-1, -2)


def test_round():
    """Round the vector values"""
    assert round(Vec2(1.1, 2.2)) == Vec2(1, 2)
    assert round(Vec2(1.1, 2.2), 1) == Vec2(1.1, 2.2)


def test_ceil():
    assert math.ceil(Vec2(1.1, 2.2)) == Vec2(2, 3)


def test_floor():
    assert math.floor(Vec2(1.1, 2.2)) == Vec2(1, 2)


def test_trunc():
    assert math.trunc(Vec2(1.1, 2.2)) == Vec2(1, 2)


def test_mod():
    assert Vec2(3, 4) % Vec2(2, 2) == Vec2(1, 0)
    assert Vec2(3, 4) % 2 == Vec2(1, 0)

    assert Vec2(5.3, 4.7) % 2 == pytest.approx(Vec2(1.3, 0.7), abs=1e-5)


def test_pow():
    assert Vec2(2, 3) ** Vec2(2, 3) == Vec2(4, 27)
    assert Vec2(2, 3) ** 2 == Vec2(4, 9)


def test_lt():
    """Compares the length of the vectors."""
    assert Vec2(1, 2) < Vec2(2, 3)

    assert Vec2(1, 2) < (2, 3)
    assert (1, 2) < Vec2(2, 3)


def test_sum():
    assert sum(Vec2(1, 2)) == 3


def test_from_polar():
    """Create a vector from polar coordinates."""
    assert Vec2.from_polar(math.radians(90)) == pytest.approx(Vec2(0, 1), abs=1e-5)
    assert Vec2.from_polar(math.radians(180)) == pytest.approx(Vec2(-1, 0), abs=1e-5)
    assert Vec2.from_polar(math.radians(270)) == pytest.approx(Vec2(0, -1), abs=1e-5)


def test_from_heading():
    """Create a vector from a heading."""
    assert Vec2.from_heading(0, length=1) == Vec2(1, 0)
    assert Vec2.from_heading(math.radians(90), length=2) == pytest.approx(Vec2(0, 2), abs=1e-5)
    assert Vec2.from_heading(math.radians(180), length=3) == pytest.approx(Vec2(-3, 0), abs=1e-5)
    assert Vec2.from_heading(math.radians(-90), length=4) == pytest.approx(Vec2(0, -4), abs=1e-5)


def test_heading():
    assert Vec2(1, 0).heading() == 0
    assert Vec2(0, 1).heading() == math.radians(90)
    assert Vec2(-1, 0).heading() == math.radians(180)
    assert Vec2(0, -1).heading() == math.radians(-90)


def test_lerp():
    """Linear interpolation between two vectors."""
    assert Vec2(1, 2).lerp(Vec2(3, 4), 0.5) == Vec2(2, 3)
    assert Vec2(1, 2).lerp(Vec2(3, 4), 0) == Vec2(1, 2)
    assert Vec2(1, 2).lerp(Vec2(3, 4), 1) == Vec2(3, 4)
    assert Vec2(1, 2).lerp((3, 4), 0.5) == Vec2(2, 3)


def test_step():
    """Step function."""
    assert Vec2(1, 1).step(Vec2(1, 1)) == Vec2(1, 1)
    assert Vec2(0, 2).step(Vec2(1, 1)) == Vec2(0, 1)
    assert Vec2(2, 0).step(Vec2(1, 1)) == Vec2(1, 0)
    assert Vec2(3, 3).step(Vec2(1, 1)) == Vec2(1, 1)


def test_reflect():
    """Reflect vector off another vector."""
    assert Vec2(1, 0). reflect(Vec2(1, 0)) == Vec2(-1, 0)
    assert Vec2(1, 0). reflect(Vec2(0, 1)) == Vec2(1, 0)
    assert Vec2(1, 0). reflect(Vec2(1, 0)) == Vec2(-1, 0)


def test_rotate():
    """Rotate vector by degrees."""
    assert Vec2(1, 0).rotate(math.radians(90)) == pytest.approx(Vec2(0, 1), abs=1e-5)
    assert Vec2(1, 0).rotate(math.radians(180)) == pytest.approx(Vec2(-1, 0), abs=1e-5)
    assert Vec2(1, 0).rotate(math.radians(270)) == pytest.approx(Vec2(0, -1), abs=1e-5)
    assert Vec2(1, 0).rotate(math.radians(360)) == pytest.approx(Vec2(1, 0), abs=1e-5)


def test_distance():
    """Calculate the distance between two vectors."""
    assert Vec2(1, 2).distance(Vec2(3, 4)) == pytest.approx(2.82843, abs=1e-5)
    assert Vec2(-1, -2).distance(Vec2(3, 4)) == pytest.approx(7.21110, abs=1e-5)


def test_normalize():
    """Normalize various vectors."""
    assert Vec2(1, 2).normalize() == pytest.approx(Vec2(0.447214, 0.894427), abs=1e-5)
    assert Vec2(2, 1).normalize() == pytest.approx(Vec2(0.894427, 0.447214), abs=1e-5)
    assert Vec2(-1, -2).normalize() == pytest.approx(Vec2(-0.447214, -0.894427), abs=1e-5)
    assert Vec2(-2, -1).normalize() == pytest.approx(Vec2(-0.894427, -0.447214), abs=1e-5)
    assert Vec2(0, 0).normalize() == Vec2(0, 0)


def test_clamp():
    """Attempt to clamp vector using all possible combinations."""
    assert Vec2(-10, 10).clamp(-10, 10) == Vec2(-10, 10)
    assert Vec2(-10, 10).clamp(-20, 5) == Vec2(-10, 5)
    assert Vec2(-10, 10).clamp(-5, 5) == Vec2(-5, 5)
    assert Vec2(-10, 10).clamp(5, 15) == Vec2(5, 10)
    assert Vec2(-10, 10).clamp(5, 5) == Vec2(5, 5)

    assert Vec2(-10, 10).clamp(Vec2(-20, -20), Vec2(20, 20)) == Vec2(-10, 10)
    assert Vec2(-10, 10).clamp(Vec2(-5, 5), Vec2(5, 5)) == Vec2(-5, 5)

    # Revert if necessary for perf
    assert Vec2(-1, 50).clamp(Vec2(0, 0), 10) == Vec2(0, 10)
    assert Vec2(-1,  50).clamp(0, Vec2(10, 10)) == Vec2(0, 10)


def test_dot():
    assert Vec2(1, 0).dot(Vec2(1, 0)) == 1  # same direction
    assert Vec2(1, 0).dot(Vec2(0, 1)) == 0  # perpendicular
    assert Vec2(1, 2).dot(Vec2(3, 4)) == 11

    assert Vec2(-1, 0).dot((-1, 0)) == 1  # same direction
    assert Vec2(-1, 0).dot((0, -1)) == 0  # perpendicular
    assert Vec2(-1, -2).dot((-3, -4)) == 11


def test_index():
    with pytest.raises(NotImplementedError):
        Vec2(0).index(0)


def test_bool():
    v = Vec2()
    assert not v

    v = Vec2(0.1, 2.0)
    assert v

    v = Vec2(-1.0, 1.0)
    assert v



""" Camera class similar to that of camera.py, this time implemented as a graphics Group.
Interface is much more simplified, with only a position and zoom implemented, but is easily
extended to add other features such as autoscroll.

    camera = CameraGroup(x=0, y=0, zoom=1)

    world_object = pyglet.some_renderable(batch=batch, group=camera)
    ui_object = pyglet.some_renderable(batch=batch)  # <-- Using the same batch here

    @window.event
    def on_draw():
        window.clear()
        batch.draw()  # Only one batch necessary

A centered camera class is also provided, where the position of the camera is the center of
the screen instead of the bottom left.

    centered_camera = CenteredCameraGroup(window, x=0, y=0, zoom=1)

Demo:

Use arrow keys to move the camera around the scene.
Note that everything in the window can be added to the same batch, as a group is used to
separate things in world space from things in "UI" space.
"""

import pyglet
from pyglet.graphics import Group
from pyglet.math import Vec2



class CameraGroup(Group):
    """ Graphics group emulating the behaviour of a camera in 2D space. """

    def __init__(self, window, x, y, zoom=1.0, order=0, parent=None):
        super().__init__(order, parent)
        self._window = window
        self.x = x
        self.y = y
        self.zoom = zoom

    @property
    def position(self) -> Vec2:
        """Query the current offset."""
        return Vec2(self.x, self.y)

    @position.setter
    def position(self, new_position: Vec2):
        """Set the scroll offset directly."""
        self.x, self.y = new_position

    def set_state(self):
        """ Apply zoom and camera offset to view matrix. """

        # Translate using the offset.
        view_matrix = self._window.view.translate(-self.x * self.zoom, -self.y * self.zoom, 0)
        # Scale by zoom level.
        view_matrix = view_matrix.scale(self.zoom, self.zoom, 1)

        self._window.view = view_matrix

    def unset_state(self):
        """ Revert zoom and camera offset from view matrix. """
        # Since this is a matrix, you will need to reverse the translate after rendering otherwise
        # it will multiply the current offset every draw update pushing it further and further away.

        # Use inverse zoom to reverse zoom
        view_matrix = self._window.view.scale(1 / self.zoom, 1 / self.zoom, 1)
        # Reverse translate.
        view_matrix = view_matrix.translate(self.x * self.zoom, self.y * self.zoom, 0)

        self._window.view = view_matrix


class CenteredCameraGroup(CameraGroup):
    """ Alternative centered camera group.

    (0, 0) will be the center of the screen, as opposed to the bottom left.
    """

    def set_state(self):
        # Translate almost the same as normal, but add the center offset
        x = -self._window.width // 2 / self.zoom + self.x
        y = -self._window.height // 2 / self.zoom + self.y

        view_matrix = self._window.view.translate((-x * self.zoom, -y * self.zoom, 0))
        view_matrix = view_matrix.scale((self.zoom, self.zoom, 1))
        self._window.view = view_matrix

    def unset_state(self):

        x = -self._window.width // 2 / self.zoom + self.x
        y = -self._window.height // 2 / self.zoom + self.y

        view_matrix = self._window.view.scale((1 / self.zoom, 1 / self.zoom, 1))
        view_matrix = view_matrix.translate((x * self.zoom, y * self.zoom, 0))
        self._window.view = view_matrix


if __name__ == "__main__":
    from pyglet.window import key

    # Create a window and a batch
    window = pyglet.window.Window(resizable=True)
    batch = pyglet.graphics.Batch()

    # Key handler for movement
    keys = key.KeyStateHandler()
    window.push_handlers(keys)

    # Use centered
    camera = CenteredCameraGroup(window, 0, 0)
    # Use un-centered
    # camera = CameraGroup(0, 0)

    # Create a scene
    rect = pyglet.shapes.Rectangle(-25, -25, 50, 50, batch=batch, group=camera)
    text = pyglet.text.Label("Text works too!", x=0, y=-50, anchor_x="center", batch=batch, group=camera)

    # Create some "UI"
    ui_text = pyglet.text.Label(
        "Simply don't add to the group to make UI static (like this)",
        anchor_y="bottom", batch=batch,
    )
    position_text = pyglet.text.Label(
        "",
        x=window.width,
        anchor_x="right", anchor_y="bottom",
        batch=batch,
    )

    @window.event
    def on_draw():
        # Draw our scene
        window.clear()
        batch.draw()

    @window.event
    def on_resize(width: float, height: float):
        # Keep position text label to the right
        position_text.x = width

    def on_update(dt: float):
        # Move camera with arrow keys
        if keys[key.UP]:
            camera.y += 50*dt
        if keys[key.DOWN]:
            camera.y -= 50*dt
        if keys[key.LEFT]:
            camera.x -= 50*dt
        if keys[key.RIGHT]:
            camera.x += 50*dt

        # Update position text label
        position_text.text = repr(round(camera.position))

    # Start the demo
    pyglet.clock.schedule(on_update)
    pyglet.app.run()



import numbers
import pytest
from threading import Timer

import pyglet
pyglet.options['debug_media'] = False

from pyglet.media.codecs import AudioFormat
from pyglet.media.synthesis import Silence

try:
    from pyglet.media.drivers.pulse import interface
except ImportError:
    interface = None


pytestmark = pytest.mark.skipif(interface is None, reason='requires PulseAudio')


@pytest.fixture
def mainloop():
    return interface.PulseAudioMainloop()


def test_mainloop_run(mainloop):
    mainloop.start()
    mainloop.delete()


def test_mainloop_lock(mainloop):
    mainloop.start()
    mainloop.lock_()
    mainloop.unlock()
    mainloop.delete()


def test_mainloop_signal(mainloop):
    mainloop.start()
    with mainloop.lock:
        mainloop.signal()
    mainloop.delete()


def test_mainloop_wait_signal(mainloop):
    mainloop.start()

    def signal():
        with mainloop.lock:
            mainloop.signal()
    t = Timer(.1, signal)
    t.start()

    with mainloop.lock:
        mainloop.wait()
    mainloop.delete()


@pytest.fixture
def context(mainloop):
    mainloop.start()
    with mainloop.lock:
        context = mainloop.create_context()
    yield context

    with context.mainloop.lock:
        context.delete()
    mainloop.delete()


def test_context_not_connected(context):
    assert context.is_ready == False
    assert context.is_failed == False
    assert context.is_terminated == False
    assert context.server == None
    assert isinstance(context.protocol_version, numbers.Integral)
    assert context.server_protocol_version == None
    assert context.is_local == None

    with context.mainloop.lock:
        context.delete()

    assert context.is_ready == False
    assert context.is_failed == False
    assert context.is_terminated == False  # Never connected, so state is not set yet
    assert context.server == None
    assert context.protocol_version == None
    assert context.server_protocol_version == None
    assert context.is_local == None


def test_context_connect(context):
    assert context.is_ready == False
    assert context.is_failed == False
    assert context.is_terminated == False
    assert context.server == None
    assert isinstance(context.protocol_version, numbers.Integral)
    assert context.server_protocol_version == None
    assert context.is_local == None

    context.connect()

    assert context.is_ready == True
    assert context.is_failed == False
    assert context.is_terminated == False
    assert isinstance(context.server, str)
    assert isinstance(context.protocol_version, numbers.Integral)
    assert isinstance(context.server_protocol_version, numbers.Integral)
    assert context.is_local == True

    with context.mainloop.lock:
        context.delete()

    assert context.is_ready == False
    assert context.is_failed == False
    assert context.is_terminated == True
    assert context.server == None
    assert context.protocol_version == None
    assert context.server_protocol_version == None
    assert context.is_local == None


@pytest.fixture
def stream(context):
    context.connect()
    audio_format = AudioFormat(1, 16, 44100)
    with context.mainloop.lock:
        stream = context.create_stream(audio_format)
    return stream


@pytest.fixture
def audio_source():
    return Silence(10.0, 44100)


@pytest.fixture
def filled_stream(stream, audio_source):
    with stream.mainloop.lock:
        stream.connect_playback()

    assert stream.is_ready
    with stream.mainloop.lock:
        writable_size = stream.get_writable_size()
        assert writable_size > 0

    nbytes = min(1024, writable_size)
    audio_data = audio_source.get_audio_data(nbytes)
    with stream.mainloop.lock:
        stream.write(audio_data.pointer, nbytes)

    assert stream.is_ready
    return stream


def test_stream_create(stream):
    assert stream.is_unconnected == True
    assert stream.is_creating == False
    assert stream.is_ready == False
    assert stream.is_failed == False
    assert stream.is_terminated == False

    with stream.mainloop.lock:
        stream.delete()

    assert stream.is_unconnected == True
    assert stream.is_creating == False
    assert stream.is_ready == False
    assert stream.is_failed == False
    assert stream.is_terminated == False


def test_stream_connect(stream):
    assert stream.is_unconnected == True
    assert stream.is_creating == False
    assert stream.is_ready == False
    assert stream.is_failed == False
    assert stream.is_terminated == False
    assert stream.index is None

    with stream.mainloop.lock:
        stream.connect_playback()

    assert stream.is_unconnected == False
    assert stream.is_creating == False
    assert stream.is_ready == True
    assert stream.is_failed == False
    assert stream.is_terminated == False
    assert isinstance(stream.index, numbers.Integral)

    with stream.mainloop.lock:
        stream.delete()

    assert stream.is_unconnected == False
    assert stream.is_creating == False
    assert stream.is_ready == False
    assert stream.is_failed == False
    assert stream.is_terminated == True



def test_stream_write(stream, audio_source):
    with stream.mainloop.lock:
        stream.connect_playback()

    assert stream.is_ready

    with stream.mainloop.lock:
        writable_size = stream.get_writable_size()
        assert writable_size > 0

    nbytes = min(1024, writable_size)
    audio_data = audio_source.get_audio_data(nbytes)
    with stream.mainloop.lock:
        written = stream.write(audio_data.pointer, nbytes)
    assert written == nbytes

    assert stream.is_ready

    with stream.mainloop.lock:
        stream.delete()

    assert stream.is_terminated



def test_stream_timing_info(filled_stream):
    with filled_stream.mainloop.lock:
        op = filled_stream.update_timing_info()
        op.wait()
    assert op.is_done
    op.delete()

    info = filled_stream.get_timing_info()
    assert info is not None


def test_stream_trigger(filled_stream):
    with filled_stream.mainloop.lock:
        op = filled_stream.trigger()
        op.wait()
    assert op.is_done
    op.delete()


def test_stream_prebuf(filled_stream):
    with filled_stream.mainloop.lock:
        op = filled_stream.prebuf()
        op.wait()
    assert op.is_done
    op.delete()


def test_stream_cork(filled_stream):
    assert filled_stream.is_corked()

    with filled_stream.mainloop.lock:
        op = filled_stream.resume()
        op.wait()
        assert op.is_done
        assert not filled_stream.is_corked()
    op.delete()

    with filled_stream.mainloop.lock:
        op = filled_stream.pause()
        op.wait()
        assert op.is_done
        assert filled_stream.is_corked()
    op.delete()


def test_stream_update_sample_rate(filled_stream):
    with filled_stream.mainloop.lock:
        op = filled_stream.update_sample_rate(44100)
        op.wait()
    assert op.is_done
    op.delete()


def test_stream_write_needed(stream, audio_source, event_loop):
    with stream.mainloop.lock:
        stream.connect_playback()

    assert stream.is_ready

    with stream.mainloop.lock:
        writable_size = stream.get_writable_size()
        assert writable_size > 0

    def on_write_needed(_stream, nbytes, _userdata):
        on_write_needed.nbytes = nbytes

    on_write_needed.nbytes = None

    def on_underflow(_stream, _userdata):
        on_underflow.underflow = True
        event_loop.interrupt_event_loop()

    on_underflow.underflow = False

    stream.set_write_callback(on_write_needed)
    stream.set_underflow_callback(on_underflow)

    audio_data = audio_source.get_audio_data(writable_size)
    with stream.mainloop.lock:
        stream.write(audio_data.pointer, audio_data.length)
    assert stream.is_ready

    with stream.mainloop.lock:
        stream.resume().wait().delete()

    event_loop.run_event_loop(duration=0.1)

    assert on_write_needed.nbytes > 0
    assert on_underflow.underflow == False

    on_write_needed.nbytes = None
    on_underflow.underflow = None

    event_loop.run_event_loop(duration=5.0)
    assert on_write_needed.nbytes > 0
    assert on_underflow.underflow == True



"""
This example shows the KeyStateHandler polling approach when it comes to catching keyboard events.
"""

import pyglet

window = pyglet.window.Window()
label = pyglet.text.Label("Press either A, Enter, or Left Arrow", 50, 50, font_size=36)


key_handler = pyglet.window.key.KeyStateHandler()
window.push_handlers(key_handler)


def update(dt):

    if key_handler[pyglet.window.key.A]:
        label.text = 'The "A" key was pressed'
    elif key_handler[pyglet.window.key.LEFT]:
        label.text = 'The "Left Arrow" key was pressed.'
    elif key_handler[pyglet.window.key.ENTER]:
        label.text = 'The "Enter" key was pressed.'
    else:
        label.text = "Press either A, Enter, or Left Arrow"


@window.event
def on_draw():
    window.clear()
    label.draw()


pyglet.clock.schedule_interval(update, 1/60)


if __name__ == '__main__':
    pyglet.app.run()



from __future__ import annotations

from ctypes import c_void_p
from typing import TYPE_CHECKING

import pyglet
from pyglet.libs.darwin.cocoapy import (
    NSApplicationDidHideNotification,
    NSApplicationDidUnhideNotification,
    NSMakeRect,
    ObjCClass,
    ObjCInstance,
    ObjCSubclass,
    PyObjectEncoding,
    appkit,
    get_selector,
    quartz,
    send_super,
)

from .systemcursor import SystemCursor

if TYPE_CHECKING:
    from . import CocoaWindow

NSNotificationCenter = ObjCClass('NSNotificationCenter')
NSApplication = ObjCClass('NSApplication')
NSNotification = ObjCClass('NSNotification')

NSBackingPropertyOldScaleFactorKey = c_void_p.in_dll(appkit, 'NSBackingPropertyOldScaleFactorKey')


class PygletDelegate_Implementation:
    PygletDelegate = ObjCSubclass('NSObject', 'PygletDelegate')

    @PygletDelegate.method(b'@' + PyObjectEncoding)
    def initWithWindow_(self, window: CocoaWindow) -> ObjCInstance | None:
        self = ObjCInstance(send_super(self, 'init'))

        if not self:
            return None

        # CocoaWindow object.
        self._window = window
        window._nswindow.setDelegate_(self)  # noqa: SLF001


        # Register delegate for hide and unhide notifications so that we
        # can dispatch the corresponding pyglet events.
        notificationCenter = NSNotificationCenter.defaultCenter()

        notificationCenter.addObserver_selector_name_object_(
            self, get_selector('applicationDidHide:'),
            NSApplicationDidHideNotification, None)

        notificationCenter.addObserver_selector_name_object_(
            self, get_selector('applicationDidUnhide:'),
            NSApplicationDidUnhideNotification, None)

        # Flag set when we pause exclusive mouse mode if window loses key status.
        self.did_pause_exclusive_mouse = False
        return self

    @PygletDelegate.method('v')
    def dealloc(self) -> None:
        # Unregister delegate from notification center.
        notificationCenter = NSNotificationCenter.defaultCenter()
        notificationCenter.removeObserver_(self)
        self._window = None
        send_super(self, 'dealloc')

    @PygletDelegate.method('v@')
    def applicationDidHide_(self, notification: NSNotification) -> None:
        self._window.dispatch_event('on_hide')

    @PygletDelegate.method('v@')
    def applicationDidUnhide_(self, notification: NSNotification) -> None:
        if self._window._mouse_exclusive and quartz.CGCursorIsVisible():
            # The cursor should be hidden, but for some reason it's not;
            # try to force the cursor to hide (without over-hiding).
            SystemCursor.unhide()
            SystemCursor.hide()
        self._window.dispatch_event('on_show')

    @PygletDelegate.method('B@')
    def windowShouldClose_(self, sender: ObjCInstance) -> bool:
        # The method is not called if [NSWindow close] was used.
        self._window.dispatch_event('on_close')
        return False

    @PygletDelegate.method('v@')
    def windowDidMove_(self, notification: NSNotification) -> None:
        x, y = self._window.get_location()
        self._window.dispatch_event('on_move', x, y)

    @PygletDelegate.method('v@')
    def windowDidBecomeKey_(self, notification: NSNotification) -> None:
        # Restore exclusive mouse mode if it was active before we lost key status.
        if self.did_pause_exclusive_mouse:
            self._window.set_exclusive_mouse(True)
            self.did_pause_exclusive_mouse = False
            self._window._nswindow.setMovable_(True)  # Mac OS 10.6  # noqa: SLF001
        # Restore previous mouse visibility settings.
        self._window.set_mouse_platform_visible()
        self._window.dispatch_event('on_activate')

    @PygletDelegate.method('v@')
    def windowDidResignKey_(self, notification: NSNotification) -> None:
        # Pause exclusive mouse mode if it is active.
        if self._window._mouse_exclusive:  # noqa: SLF001
            self._window.set_exclusive_mouse(False)
            self.did_pause_exclusive_mouse = True
            # We need to prevent the window from being unintentionally dragged
            # (by the call to set_mouse_position in set_exclusive_mouse) when
            # the window is reactivated by clicking on its title bar.
            self._window._nswindow.setMovable_(False)  # Mac OS X 10.6  # noqa: SLF001
        # Make sure that cursor is visible.
        self._window.set_mouse_platform_visible(True)
        self._window.dispatch_event('on_deactivate')

    @PygletDelegate.method('v@')
    def windowDidMiniaturize_(self, notification: NSNotification) -> None:
        self._window.dispatch_event('on_hide')

    @PygletDelegate.method('v@')
    def windowDidDeminiaturize_(self, notification: NSNotification) -> None:
        if self._window._mouse_exclusive and quartz.CGCursorIsVisible():  # noqa: SLF001
            # The cursor should be hidden, but for some reason it's not;
            # try to force the cursor to hide (without over-hiding).
            SystemCursor.unhide()
            SystemCursor.hide()
        self._window.dispatch_event('on_show')

    @PygletDelegate.method('v@')
    def windowDidExpose_(self, notification: NSNotification) -> None:
        self._window.dispatch_event('on_expose')

    @PygletDelegate.method('v@')
    def terminate_(self, sender: ObjCInstance) -> None:
        NSApp = NSApplication.sharedApplication()
        NSApp.terminate_(self)

    @PygletDelegate.method('B@')
    def validateMenuItem_(self, menuitem: ObjCInstance) -> bool:
        # Disable quitting with command-q when in keyboard exclusive mode.
        if menuitem.action() == get_selector('terminate:'):
            return not self._window._keyboard_exclusive  # noqa: SLF001
        return True

    @PygletDelegate.method('v@')
    def windowDidChangeBackingProperties_(self, notification):
        if not self._window._shadow:
            user_info = notification.userInfo()
            old_scale_value = user_info.objectForKey_(NSBackingPropertyOldScaleFactorKey)

            old_scale = old_scale_value.doubleValue()
            new_scale = self._window._nswindow.backingScaleFactor()
            if old_scale != new_scale:
                self._window.switch_to()

                currentFrame = self._window._nswindow.frame()

                if pyglet.options.dpi_scaling == "real":
                    screen_scale = new_scale
                    w, h = self._window.get_requested_size()
                    width, height = int(w / screen_scale), int(h / screen_scale)

                    # Force Window back to correct size.
                    self._window._set_frame_size(width, height)
                else:
                    # MacOS seems to cache the state of the window size, even between different DPI scales/monitors.
                    # This means that the screen will refuse to refresh until we resize the window to a different size.
                    # Force a refresh by setting a temporary frame, then forcing it back.
                    if pyglet.options.dpi_scaling == "scaled":
                        tempRect = NSMakeRect(currentFrame.origin.x, currentFrame.origin.y,
                                              currentFrame.size.width + 1, currentFrame.size.height + 1)
                        # TODO: Add variable to ignore the next two on-resize events?
                        self._window._nswindow.setFrame_display_(tempRect, True)
                        self._window._nswindow.setFrame_display_(currentFrame, True)

                self._window.context.update_geometry()
                self._window.dispatch_event("_on_internal_scale", new_scale, self._window._get_dpi_desc())



PygletDelegate = ObjCClass('PygletDelegate')



"""Event dispatch framework.

All objects that produce events in pyglet implement :py:class:`~pyglet.event.EventDispatcher`,
providing a consistent interface for registering and manipulating event
handlers.  A commonly used event dispatcher is `pyglet.window.Window`.

Event types
===========

For each event dispatcher there is a set of events that it dispatches; these
correspond with the type of event handlers you can attach.  Event types are
identified by their name, for example, ''on_resize''.  If you are creating a
new class which implements :py:class:`~pyglet.event.EventDispatcher`, you must call
`EventDispatcher.register_event_type` for each event type.

Attaching event handlers
========================

An event handler is simply a function or method.  You can attach an event
handler by setting the appropriate function on the instance::

    def on_resize(width, height):
        # ...
    dispatcher.on_resize = on_resize

There is also a convenience decorator that reduces typing::

    @dispatcher.event
    def on_resize(width, height):
        # ...

You may prefer to subclass and override the event handlers instead::

    class MyDispatcher(DispatcherClass):
        def on_resize(self, width, height):
            # ...

Event handler stack
===================

When attaching an event handler to a dispatcher using the above methods, it
replaces any existing handler (causing the original handler to no longer be
called).  Each dispatcher maintains a stack of event handlers, allowing you to
insert an event handler "above" the existing one rather than replacing it.

There are two main use cases for "pushing" event handlers:

* Temporarily intercepting the events coming from the dispatcher by pushing a
  custom set of handlers onto the dispatcher, then later "popping" them all
  off at once.
* Creating "chains" of event handlers, where the event propagates from the
  top-most (most recently added) handler to the bottom, until a handler
  takes care of it.

Use `EventDispatcher.push_handlers` to create a new level in the stack and
attach handlers to it.  You can push several handlers at once::

    dispatcher.push_handlers(on_resize, on_key_press)

If your function handlers have different names to the events they handle, use
keyword arguments::

    dispatcher.push_handlers(on_resize=my_resize, on_key_press=my_key_press)

After an event handler has processed an event, it is passed on to the
next-lowest event handler, unless the handler returns `EVENT_HANDLED`, which
prevents further propagation.

To remove all handlers on the top stack level, use
`EventDispatcher.pop_handlers`.

Note that any handlers pushed onto the stack have precedence over the
handlers set directly on the instance (for example, using the methods
described in the previous section), regardless of when they were set.
For example, handler ``foo`` is called before handler ``bar`` in the following
example::

    dispatcher.push_handlers(on_resize=foo)
    dispatcher.on_resize = bar

Dispatching events
==================

pyglet uses a single-threaded model for all application code.  Event
handlers are only ever invoked as a result of calling
EventDispatcher.dispatch_events`.

It is up to the specific event dispatcher to queue relevant events until they
can be dispatched, at which point the handlers are called in the order the
events were originally generated.

This implies that your application runs with a main loop that continuously
updates the application state and checks for new events::

    while True:
        dispatcher.dispatch_events()
        # ... additional per-frame processing

Not all event dispatchers require the call to ``dispatch_events``; check with
the particular class documentation.

.. note::

    In order to prevent issues with garbage collection, the
    :py:class:`~pyglet.event.EventDispatcher` class only holds weak
    references to pushed event handlers. That means the following example
    will not work, because the pushed object will fall out of scope and be
    collected::

        dispatcher.push_handlers(MyHandlerClass())

    Instead, you must make sure to keep a reference to the object before pushing
    it. For example::

        my_handler_instance = MyHandlerClass()
        dispatcher.push_handlers(my_handler_instance)

"""
from __future__ import annotations

import inspect
import os.path

from functools import partial
from typing import TYPE_CHECKING, Literal, Union
from weakref import WeakMethod

import pyglet

if TYPE_CHECKING:
    from typing import Any, Callable, Generator


EVENT_HANDLED = True
EVENT_UNHANDLED = None

EVENT_HANDLE_STATE = Union[Literal[True], None]


class EventException(Exception):  # noqa: N818
    """An exception raised when an event handler could not be attached."""


class EventDispatcher:
    """Generic event dispatcher interface.

    See the module docstring for usage.
    """
    event_types: list
    # Placeholder empty stack; real stack is created only if needed
    _event_stack: tuple | list = ()

    @classmethod
    def register_event_type(cls: type[object], name: str) -> str:
        """Register an event type with the dispatcher.

        Before dispatching events, they must first be registered by name.
        Registering event types allows the dispatcher to validate event
        handler names as they are attached, and to search attached objects
        for suitable handlers.
        """
        if not hasattr(cls, 'event_types'):
            cls.event_types = []  # type: ignore reportAttributeAccessIssue
        cls.event_types.append(name)  # type: ignore reportAttributeAccessIssue
        return name

    def push_handlers(self, *args: Any, **kwargs: Any) -> None:
        """Push a new level onto the handler stack, and add 0 or more handlers.

        This method first pushes a new level to the top of the handler stack.
        It then attaches any handlers that were passed to this new level.

        If keyword arguments are given, they name the event type to attach.
        Otherwise, a callable's ``__name__`` attribute will be used. Any
        other object may also be specified, in which case it will be searched
        for callables with event names.
        """
        # Create event stack if necessary
        if type(self._event_stack) is tuple:
            self._event_stack = []

        # Place dict full of new handlers at beginning of stack
        self._event_stack.insert(0, {})  # type: ignore reportAttributeAccessIssue
        self.set_handlers(*args, **kwargs)

    def _get_handlers(self, args: list, kwargs: dict) -> Generator[tuple[str, Callable], None, None]:
        """Implement handler matching on arguments for set_handlers and remove_handlers."""
        for obj in args:
            if inspect.isroutine(obj):
                # Single magically named function
                name: str = obj.__name__
                if name not in self.event_types:
                    msg = f'Unknown event "{name}"'
                    raise EventException(msg)
                if inspect.ismethod(obj):
                    yield name, WeakMethod(obj, partial(self._remove_handler, name))
                else:
                    yield name, obj
            else:
                # Single instance with magically named methods
                for name in dir(obj):
                    if name in self.event_types:
                        meth = getattr(obj, name)
                        yield name, WeakMethod(meth, partial(self._remove_handler, name))

        for name, handler in kwargs.items():
            # Function for handling given event (no magic)
            if name not in self.event_types:
                msg = f'Unknown event "{name}"'
                raise EventException(msg)
            if inspect.ismethod(handler):
                yield name, WeakMethod(handler, partial(self._remove_handler, name))
            else:
                yield name, handler

    def set_handlers(self, *args: Any, **kwargs: Any) -> None:
        """Attach one or more event handlers to the top level of the handler stack.

        See :py:meth:`~pyglet.event.EventDispatcher.push_handlers` for the accepted
        argument types.
        """
        # Create event stack if necessary
        if type(self._event_stack) is tuple:
            self._event_stack = [{}]

        for name, handler in self._get_handlers(args, kwargs):
            self.set_handler(name, handler)

    def set_handler(self, name: str, handler: Callable) -> None:
        """Attach a single event handler."""
        # Create event stack if necessary
        if type(self._event_stack) is tuple:
            self._event_stack = [{}]

        self._event_stack[0][name] = handler

    def pop_handlers(self) -> None:
        """Pop the top level of event handlers off the stack."""
        assert self._event_stack, 'No handlers pushed'

        del self._event_stack[0]

    def remove_handlers(self, *args: Any, **kwargs: Any) -> None:
        """Remove event handlers from the event stack.

        See :py:meth:`~pyglet.event.EventDispatcher.push_handlers` for the
        accepted argument types. All handlers are removed from the first stack
        frame that contains any of the given handlers. No error is raised if
        any handler does not appear in that frame, or if no stack frame
        contains any of the given handlers.

        If the stack frame is empty after removing the handlers, it is
        removed from the stack.  Note that this interferes with the expected
        symmetry of :py:meth:`~pyglet.event.EventDispatcher.push_handlers` and
        :py:meth:`~pyglet.event.EventDispatcher.pop_handlers`.
        """
        handlers = list(self._get_handlers(args, kwargs))

        # Find the first stack frame containing any of the handlers
        def find_frame() -> dict | None:
            for _frame in self._event_stack:
                for _name, _handler in handlers:
                    if _name not in _frame:
                        continue
                    if _frame[_name] == _handler:
                        return _frame
            return None

        frame = find_frame()

        # No frame matched; no error.
        if not frame:
            return

        # Remove each handler from the frame.
        for name, handler in handlers:
            try:
                if frame[name] == handler:
                    del frame[name]
            except KeyError:  # noqa: PERF203
                pass

        # Remove the frame if it's empty.
        if not frame:
            self._event_stack.remove(frame)

    def remove_handler(self, name: str, handler: Callable) -> None:
        """Remove a single event handler.

        The given event handler is removed from the first handler stack frame
        it appears in.  The handler must be the exact same callable as passed
        to `set_handler`, `set_handlers` or
        :py:meth:`~pyglet.event.EventDispatcher.push_handlers`; and the name
        must match the event type it is bound to.

        No error is raised if the event handler is not set.
        """
        for frame in self._event_stack:
            try:
                if frame[name] == handler:
                    del frame[name]
                    break
            except KeyError:
                pass

    def _remove_handler(self, name: str, handler: Callable) -> None:
        """Used internally to remove all handler instances for the given event name.

        This is normally called from a dead ``WeakMethod`` to remove itself from the
        event stack.
        """
        # Iterate over a copy as we might mutate the list
        for frame in list(self._event_stack):

            if name in frame:
                try:
                    if frame[name] == handler:
                        del frame[name]
                        if not frame:
                            self._event_stack.remove(frame)
                except TypeError:
                    # weakref is already dead
                    pass

    def dispatch_event(self, event_type: str, *args: Any) -> bool | None:
        """Dispatch an event to the attached event handlers.

        The event is propagated to all registered event handlers
        in the stack, starting and the top and going down. If any
        registered event handler returns ``EVENT_HANDLED``, no further
        handlers down the stack will receive this event.

        This method has several possible return values. If any event
        handler has returned ``EVENT_HANDLED``, then this method will
        also return ``EVENT_HANDLED``. If not, this method will return
        ``EVENT_UNHANDLED``. If there were no events registered to
        receive this event, ``False`` is returned.

        Returns:
            ``EVENT_HANDLED`` if any event handler returned ``EVENT_HANDLED``;
            ``EVENT_UNHANDLED`` if one or more event handlers were invoked
            without any of them returning `EVENT_HANDLED`; ``False`` if no
            event handlers were registered.
        """
        assert hasattr(self, 'event_types'), (
            "No events registered on this EventDispatcher. "
            "You need to register events with the class method "
            "EventDispatcher.register_event_type('event_name')."
        )
        assert event_type in self.event_types, f"{event_type} not found in {self}.event_types == {self.event_types}"

        invoked = False

        # Search handler stack for matching event handlers
        for frame in list(self._event_stack):
            handler = frame.get(event_type, None)
            if not handler:
                continue
            if isinstance(handler, WeakMethod):
                handler = handler()
                assert handler is not None
            try:
                invoked = True
                if handler(*args):
                    return EVENT_HANDLED
            except TypeError as exception:
                self._raise_dispatch_exception(event_type, args, handler, exception)

        # Check instance for an event handler
        try:
            if getattr(self, event_type)(*args):
                return EVENT_HANDLED
        except AttributeError as e:
            event_op = getattr(self, event_type, None)
            if callable(event_op):
                raise e
        except TypeError as exception:
            self._raise_dispatch_exception(event_type, args, getattr(self, event_type), exception)
        else:
            invoked = True

        if invoked:
            return EVENT_UNHANDLED

        return False

    def post_event(self, event_type: str, *args: Any) -> bool | None:
        """Post an event to the main application thread.

        Unlike the :py:meth:`~pyglet.event.EventDispatcher.dispatch_event`
        method, this method does not dispatch events directly. Instead, it
        hands off the dispatch call to the main application thread. This
        ensures that any event handlers are also executed in the main thread.

        This method aliases :py:meth:`~pyglet.app.PlatformEventLoop.post_event`,
        which can be seen for more information on behavior.
        """
        pyglet.app.platform_event_loop.post_event(self, event_type, *args)

    def _raise_dispatch_exception(self, event_type: str, args: Any, handler: Callable, exception: Exception) -> None:
        # A common problem in applications is having the wrong number of
        # arguments in an event handler.  This is caught as a TypeError in
        # dispatch_event but the error message is obfuscated.
        #
        # Here we check if there is indeed a mismatch in argument count,
        # and construct a more useful exception message if so.  If this method
        # doesn't find a problem with the number of arguments, the error
        # is re-raised as if we weren't here.

        n_args = len(args)

        # Inspect the handler
        argspecs = inspect.getfullargspec(handler)
        handler_args = argspecs.args
        handler_varargs = argspecs.varargs
        handler_defaults = argspecs.defaults

        n_handler_args = len(handler_args)

        # Remove "self" arg from handler if it's a bound method
        if inspect.ismethod(handler) and handler.__self__:
            n_handler_args -= 1

        # Allow *args varargs to overspecify arguments
        if handler_varargs:
            n_handler_args = max(n_handler_args, n_args)

        # Allow default values to overspecify arguments
        if handler_defaults and n_handler_args > n_args >= n_handler_args - len(handler_defaults):
            n_handler_args = n_args

        # Construct a more informative message
        if n_handler_args != n_args:
            if inspect.isfunction(handler) or inspect.ismethod(handler):
                _, filename = os.path.split(handler.__code__.co_filename)
                descr = f"'{handler.__name__}' in {filename}:{handler.__code__.co_firstlineno}"
            else:
                descr = repr(handler)

            if 'self' in handler_args:
                handler_args.remove('self')

            caller_name = f"{self.__class__.__name__}.{event_type}"
            msg = (f"The '{caller_name}' event was dispatched with {n_args} arguments:  {list(args)},\n"
                   f"but your handler {descr} is written to expect {n_handler_args} arguments: {handler_args}")
            raise TypeError(msg)

        raise exception

    def _dump_handlers(self) -> None:

        for level, handlers in enumerate(self._event_stack):
            print(f"level: {level}")

            for event_type, handler in handlers.items():
                print(f" - '{event_type}': {handler}")

    # Decorator

    def event(self, *args: Any) -> Callable:
        """Function decorator for an event handler.

        If the function or method name matches the event name,
        the decorator can be added without arguments. Likewise,
        if the name does not match, you can provide the target
        event name by passing it as an argument.

        Name matches::

            win = window.Window()

            @win.event
            def on_resize(self, width, height):
                # ...

        Name does not match::

            @win.event('on_resize')
            def foo(self, width, height):
                # ...

        """
        # @window.event()
        if len(args) == 0:

            def decorator(function: Callable) -> Callable:
                func_name = function.__name__
                self.set_handler(func_name, function)
                return function

            return decorator

        # @window.event
        if inspect.isroutine(args[0]):
            func = args[0]
            name = func.__name__
            self.set_handler(name, func)
            return args[0]

        # @window.event('on_resize')
        if isinstance(args[0], str):
            name = args[0]

            def decorator(function: Callable) -> Callable:
                self.set_handler(name, function)
                return function

            return decorator

        msg = "Argument must be the name of the event as a `str`."
        raise TypeError(msg)



"""Generate gl.py and gl_compat.py
See : tools/requirements.txt.

We are using the opengl-registry project to extract this information from
https://raw.githubusercontent.com/KhronosGroup/OpenGL-Registry/master/xml/gl.xml

A local version gl.xml can also be used.

Usage:

# Fetch gl.xml from Khronos Github repo
python gengl.py
python gengl.py --source url

# Use local gl.xml
python gengl.py --source local
"""  # noqa: D205
import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import TextIO, Iterable

from opengl_registry import Registry, RegistryReader

REPO_ROOT = Path(__file__).parent.parent.resolve()
DEST_PATH = REPO_ROOT / "pyglet" / "gl"


def main() -> None:  # noqa: D103
    values = parse_args(sys.argv[1:])
    if values.source == "url":
        # Fetch gl.xml from Khronos Github repo
        reader = RegistryReader.from_url()
    else:
        # Use the local gl.xml file
        reader = RegistryReader.from_file(Path(REPO_ROOT / "tools" / "gl.xml"))

    registry = reader.read()

    # OpenGL extensions we want to include
    extensions = [
        "GL_ARB_multisample",
        "EXT_texture_compression_s3tc",  # For pyglet.image.codecs.dds
        "GL_EXT_framebuffer_object",  # Needed for GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS
        "GL_ARB_bindless_texture",
        "GL_ARB_gpu_shader_int64",
        "GL_NV_mesh_shader",
    ]

    core_profile = registry.get_profile(
        api="gl",
        profile="core",
        version="4.6",
        extensions=extensions,
    )
    compat_profile = registry.get_profile(
        api="gl",
        profile="compat",
        version="4.6",
        extensions=extensions,
    )
    # es_profile = registry.get_profile(
    #     api="gles2",
    #     profile="core",
    #     version="3.2",
    #     extensions=extensions,
    # )

    core_writer = PygletGLWriter(registry=core_profile, out_module=DEST_PATH / "gl")
    core_writer.run()
    compat_writer = PygletGLWriter(registry=compat_profile, out_module=DEST_PATH / "gl_compat")
    compat_writer.run()
    # es_writer = PygletGLWriter(registry=es_profile, out_file=DEST_PATH / "gl_es")
    # es_writer.run()


def parse_args(args: str) -> Namespace:  # noqa: D103
    parser = ArgumentParser()
    parser.add_argument("--source", choices=["local", "url"], default="url")
    return parser.parse_args(args)


class PygletGLWriter:
    """Writes gl.py, gl_compat.py, and gl_es.py."""

    # All gl types manually matched to ctypes.
    # Inspect registry.types
    types = {  # noqa: RUF012
        "GLenum": "c_uint",
        "GLboolean": "c_ubyte",
        "GLbitfield": "c_uint",
        "GLvoid": "None",
        "GLbyte": "c_byte",
        "GLubyte": "c_ubyte",
        "GLshort": "c_short",
        "GLushort": "c_ushort",
        "GLint": "c_int",
        "GLuint": "c_uint",
        "GLclampx": "c_uint",
        "GLsizei": "c_int",
        "GLfloat": "c_float",
        "GLclampf": "c_float",
        "GLdouble": "c_double",
        "GLclampd": "c_double",
        "GLchar": "c_char",
        "GLintptr": "c_ptrdiff_t",
        "GLsizeiptr": "c_ptrdiff_t",
        "GLint64": "c_int64",
        "GLuint64": "c_uint64",
        "GLuint64EXT": "c_uint64",
        "GLsync": "POINTER(struct___GLsync)",
        "GLDEBUGPROC": "CFUNCTYPE(None, GLenum, GLenum, GLuint, GLenum, GLsizei, POINTER(GLchar), POINTER(GLvoid))",
    }
    # All gl types matched to python types
    pythontypes = {
        "GLenum": "int",
        "GLboolean": "int",
        "GLbitfield": "int",
        "GLvoid": "None",
        "GLbyte": "bytes",
        "GLubyte": "int",
        "GLshort": "int",
        "GLushort": "int",
        "GLint": "int",
        "GLuint": "int",
        "GLclampx": "int",
        "GLsizei": "int",
        "GLfloat": "float",
        "GLclampf": "float",
        "GLdouble": "float",
        "GLclampd": "float",
        "GLchar": "bytes",
        # "GLintptr": "c_ptrdiff_t",
        # "GLsizeiptr": "c_ptrdiff_t",
        "GLint64": "int",
        "GLuint64": "int",
        "GLuint64EXT": "int",
    }
    exclude_commands = set()

    def __init__(self, *, registry: Registry, out_module: Path):
        self._registry = registry
        self._out_module = out_module
        self._out = None
        self._stub = None
        self._all = []  # Entries for __all__
        self._commands = []

    def run(self):
        """Write the file and close"""
        with open(self._out_module.with_suffix(".py"), mode='w') as out:
            self.write_template(out, REPO_ROOT / "tools" / "gl.template")
            self.write_types(out)
            self.write_enums(out)
            self.write_commands(out)
            self.write_footer(out)

        with open(self._out_module.with_suffix(".pyi"), mode='w') as stub:
            self.write_template(stub, REPO_ROOT / "tools" / "gl_stub.template")
            self.write_types(stub)
            self.write_enum_stubs(stub)
            self.write_command_stubs(stub)

    def write_lines(self, fp: TextIO, lines: Iterable[str]) -> None:
        """Write one or several lines to the out file."""
        for line in lines:
            fp.write(line)
            fp.write("\n")

    def write_template(self, fp: TextIO, template: Path) -> None:
        """Write the header."""
        with open(template) as fd:
            fp.write(fd.read())

    def write_types(self, fp: TextIO) -> None:
        """Write all types."""
        self.write_lines(fp, ["# GL type definitions"])
        self.write_lines(fp, [f"{k} = {v}" for k, v in self.types.items()])
        self.write_lines(fp, [""])
        self._all.extend(self.types.keys())

    def write_enums(self, fp: TextIO) -> None:
        """Write all enums."""
        self.write_lines(fp, ["# GL enumerant (token) definitions"])
        self.write_lines(fp, [
            f"{e.name} = {e.value_int}"
            for e in sorted(self._registry.enums.values())
        ])
        self.write_lines(fp, [""])
        self._all.extend(self._registry.enums.keys())

    def write_enum_stubs(self, fp: TextIO) -> None:
        """Write type annotations for all enums."""
        self.write_lines(fp, ["# GL enumerant (token) definitions"])
        self.write_lines(fp, [
            f"{e.name}: int"  # assume all enums values are integers
            for e in sorted(self._registry.enums.values())
        ])
        self.write_lines(fp, [""])

    def write_commands(self, fp: TextIO) -> None:
        """Write all commands."""
        self.write_lines(fp, ["# GL command definitions"])

        # _link_function params : name, restype, argtypes, requires=None, suggestions=None
        for cmd in sorted(self._registry.commands.values()):
            if cmd.name in self.exclude_commands:
                continue

            # Return type: If the function returns a pointer type ...
            if "*" in cmd.proto:
                restype = f"POINTER({cmd.ptype})"
            else:
                restype = cmd.ptype or "None"

            # Arguments can be pointer and pointer-pointer
            arguments = []
            for param in cmd.params:
                # print(cmd.name, param.name, param.ptype, "|", param.value)

                # Detect void pointers. They don't have a ptype set
                if "void" in param.value:
                    arguments.append("POINTER(GLvoid)")
                else:
                    # Ensure we actually know what the type is
                    if not self.types.get(param.ptype):
                        raise ValueError(f"ptype {param.ptype} not a known type")
                    # Handle pointer-pointer and pointers: *, **, *const*
                    if param.value.count("*") == 2:
                        arguments.append(f"POINTER(POINTER({param.ptype}))")
                    elif param.value.count("*") == 1:
                        arguments.append(f"POINTER({param.ptype})")
                    else:
                        arguments.append(param.ptype)

            argtypes = ", ".join(arguments)
            requires = f"OpenGL {cmd.requires}" if cmd.requires else "None"
            # NOTE: PROCs are optional
            # proc_name = f"PFN{cmd.name.upper()}PROC"

            self.write_lines(fp, [
                f"{cmd.name} = _link_function('{cmd.name}', {restype}, [{argtypes}], requires='{requires}')",
                # f"{proc_name} = CFUNCTYPE({restype}, {argtypes})",
            ])
            self._all.append(cmd.name)

        self.write_lines(fp, [""])

    def write_footer(self, fp: TextIO) -> None:
        """Write __all__ section."""
        self.write_lines(fp, [
            "",
            "__all__ = [",
            *[f"    '{name}'," for name in self._all],
            "]",
        ])

    def write_command_stubs(self, fp: TextIO) -> None:
        """Write type annotations for all commands."""
        self.write_lines(fp, ["# GL command definitions"])

        # _link_function params : name, restype, argtypes, requires=None, suggestions=None
        for cmd in sorted(self._registry.commands.values()):
            if cmd.name in self.exclude_commands:
                continue

            # Return type: If the function returns a pointer type ...
            if "*" in cmd.proto:
                restype = f"_Pointer[{cmd.ptype}]"
            else:
                restype = cmd.ptype or "None"

            # Arguments can be pointer and pointer-pointer
            arguments = []
            names = []
            for param in cmd.params:
                # print(cmd.name, param.name, param.ptype, "|", param.value)
                names.append(param.name)

                # Detect void pointers. They don't have a ptype set
                if "void" in param.value:
                    # The exact types which are valid is hard to determine since
                    # ctypes automatically converts arguments to the required type, so allow Any
                    arguments.append("_Pointer[GLvoid] | Any")
                else:
                    # Ensure we actually know what the type is
                    if not self.types.get(param.ptype):
                        raise ValueError(f"ptype {param.ptype} not a known type")
                    # Handle pointer-pointer and pointers: *, **, *const*
                    if param.value.count("*") == 2:
                        arguments.append(f"_Pointer[_Pointer[{param.ptype}]] | Any")
                    elif param.value.count("*") == 1:
                        arguments.append(f"_Pointer[{param.ptype}] | Any")
                    else:
                        arguments.append(param.ptype)

            # Arguments can be pointer and pointer-pointer
            argannotations = ", ".join(
                f"{name}: {f'{arg} | {self.pythontypes[arg]}' if arg in self.pythontypes else arg}" for name, arg in
                zip(names, arguments))

            self.write_lines(fp, [
                f"def {cmd.name}({argannotations}) -> {self.pythontypes.get(restype, restype)}: ..."
            ])


if __name__ == "__main__":
    main()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Sequence
    from pyglet.model import Model
    from pyglet.graphics import Batch, Group


class Scene(ABC):
    """A high level container for one or more Node objects."""

    def __init__(self, nodes: list[Node] | None = None) -> None:
        self.nodes = nodes or []

    # TODO: test and implement:
    # def __iter__(self):
    #     """Iterate over all Nodes and their children (if existing)."""
    #     for top_node in self.nodes:
    #         for node in top_node:
    #             yield node

    def create_models(self, batch: Batch, group: Group | None = None) -> list[Model]:
        """TBD"""
        raise NotImplementedError(f"{self.__class__.__name__} does not implement this method.")

    def __repr__(self):
        return f"{self.__class__.__name__}(nodes={self.nodes})"


class Node:
    """Container for one or more sub-objects, such as Meshes."""
    def __init__(self, nodes: list[Node] | None = None, meshes: list[Mesh] | None = None,
                 skins: list[Skin] | None = None, cameras: list[Camera] | None = None) -> None:
        self.nodes = nodes or []
        self.meshes = meshes or []
        self.skins = skins or []
        self.cameras = cameras or []

    # TODO: test and implement:
    # def __iter__(self):
    #     yield self
    #     for child_node in self.nodes:
    #         yield child_node

    def __repr__(self):
        return (f"Node(nested_nodes={len(self.nodes)}, meshes={len(self.meshes)},"
                f"skins={len(self.skins)}, cameras={len(self.cameras)})")


class Mesh:
    """Object containing vertex and related data."""
    def __init__(self, primitives: list[Primitive] | None = None, name: str = "unknown") -> None:
        self.primitives = primitives or []
        self.name = name

    def __repr__(self):
        return f"Mesh(name='{self.name}', primitive_count={len(self.primitives)})"


class Attribute:
    def __init__(self, name: str, fmt: str, attr_type, count, array):
        self.name = name
        self.fmt = fmt
        self.type = attr_type
        self.count = count
        self.array = array

    def __repr__(self):
        return f"{self.__class__.__name__}(name='{self.name}', fmt={self.fmt}, type={self.type}, count={self.count})"


class Primitive:
    """Geometry to be rendered, with optional material."""
    def __init__(self, attributes: list[Attribute], indices: Sequence[int] | None,
                 mode: int, material: Material | None = None) -> None:
        self.attributes = attributes
        self.indices = indices
        self.mode = mode
        self.material = material

    def __repr__(self):
        return f"Primitive(attributes={list(self.attributes)}, mode={self.mode})"


class Material(ABC):
    """Base class for Material types"""


class SimpleMaterial(Material):
    def __init__(self, name: str = "default",
                 diffuse: Sequence[float] = (0.8, 0.8, 0.8, 1.0),
                 ambient: Sequence[float] = (0.2, 0.2, 0.2, 1.0),
                 specular: Sequence[float] = (0.0, 0.0, 0.0, 1.0),
                 emission: Sequence[float] = (0.0, 0.0, 0.0, 1.0),
                 shininess: float = 20,
                 texture_name: str | None = None) -> None:

        self.name = name
        self.diffuse = diffuse
        self.ambient = ambient
        self.specular = specular
        self.emission = emission
        self.shininess = shininess
        self.texture_name = texture_name

    def __repr__(self):
        return f"Material(name='{self.name}', texture='{self.texture_name}'"


class PBRMaterial(Material):
    def __init__(self):
        # TODO: implement this class
        pass


class Camera:
    def __init__(self, camera_type: str, aspect: float, yfov: float,
                 xmag: float, ymag: float, zfar: float, znear: float) -> None:
        self.type = camera_type
        # Perspective
        self.aspect_ratio = aspect
        self.yfov = yfov
        # Orthographic
        self.xmag = xmag
        self.ymag = ymag
        # Shared
        self.zfar = zfar
        self.znear = znear

    def __repr__(self):
        return f"Camera(type='{self.type}')"


class Skin:
    def __init__(self) -> None:
        # TODO: implement this class
        pass



#
# pyglet documentation build configuration file.
#
# This file is execfile()d with the current directory set to its containing dir.
import os
import sys
import time
import typing
import datetime

from typing import TypeVar

import sphinx_autodoc_typehints

# -- Extensions to the  Napoleon GoogleDocstring class ---------------------
from sphinx.ext.napoleon.docstring import GoogleDocstring
from sphinx.parsers import RSTParser


def parse_attributes_section(self, section):
    # Combination of _format_fields and _parse_attributes_section to get type hints loaded properly using
    # the theme and getting the custom name instead of Specifying just 'Variables'
    field_type = "Class Variables"
    fields = self._consume_fields()

    field_type = ":%s:" % field_type.strip()
    padding = " " * len(field_type)
    multi = len(fields) > 1
    lines: list[str] = []
    for _name, _type, _desc in fields:
        if not _type:
            _type = self._lookup_annotation(_name)
            if _name in self._annotations:
                _type = sphinx_autodoc_typehints.add_type_css_class(
                    sphinx_autodoc_typehints.format_annotation(self._annotations[_name], self._app.config))

        field = self._format_field(_name, _type, _desc)
        if multi:
            if lines:
                lines.extend(self._format_block(padding + " * ", field))
            else:
                lines.extend(self._format_block(field_type + " * ", field))
        else:
            lines.extend(self._format_block(field_type + " ", field))
    if lines and lines[-1]:
        lines.append("")
    return lines


GoogleDocstring._parse_attributes_section = parse_attributes_section


# Unpatch... the patch.
def _revert_patch():
    pass


sphinx_autodoc_typehints.patches._patch_google_docstring_lookup_annotation = _revert_patch


# Patch to fix our RST insertion issues.
class _RstSnippetParser(RSTParser):
    @staticmethod
    def decorate(_content) -> None:
        """Override to skip processing rst_epilog/rst_prolog for typing."""


sphinx_autodoc_typehints.parser.RSTParser = _RstSnippetParser


def write_build(data, filename):
    with open(os.path.join("internal", filename), "w") as f:
        f.write(".. list-table::\n")
        f.write("   :widths: 50 50\n")
        f.write("\n")
        for var, val in data:
            f.write("   * - " + var + "\n     - " + val + "\n")


sys.is_pyglet_doc_run = True

document_modules = ["pyglet"]

# Patched extensions base path.
sys.path.insert(0, os.path.abspath("."))

# import the pyglet package.
sys.path.insert(0, os.path.abspath(".."))

try:
    import pyglet

    print(f"Generating pyglet {pyglet.version} Documentation")
except ImportError:
    print("ERROR: pyglet not found")
    sys.exit(1)

# -- PYGLET DOCUMENTATION CONFIGURATION ----------------------------------------


# Set up substitutions that can be referenced later.
# IMPORTANT:
# Substitutions are moody and NEED control-like characters to be escaped
# after them. For example, in |min_python_version|\+ , escaping the plus
# sign is crucial to avoiding incomprehensible errors.
# For convenience, |min_python_version_fancy_str| is defined below to
# avoid having to deal with syntax errors.
# Also, please note that there does not appear to be a good way to use
# substitutions within link text by default.
rst_epilog = rf"""
.. |min_python_version| replace:: {pyglet.MIN_PYTHON_VERSION_STR}
.. |min_python_version_package_name| replace:: ``python{pyglet.MIN_PYTHON_VERSION_STR}``
.. |min_python_version_fancy_str| replace:: Python {pyglet.MIN_PYTHON_VERSION_STR}\+
"""

implementations = ["cocoa", "win32", "xlib"]

# For each module, a list of submodules that should not be imported.
# If value is None, do not try to import any submodule.
skip_modules = {"pyglet": {
                     "pyglet.lib": None,
                     "pyglet.libs": None,
                     "pyglet.app": implementations,
                     "pyglet.display": implementations + ["xlib_vidmoderestore"],
                     "pyglet.extlibs": None,
                     "pyglet.font": ["quartz",
                                     "win32",
                                     "freetype",
                                     "freetype_lib",
                                     "fontconfig",
                                     "win32query"],
                     "pyglet.input": ["darwin_hid",
                                      "directinput",
                                      "evdev",
                                      "wintab",
                                      "x11_xinput", "x11_xinput_tablet"],
                     "pyglet.image.codecs": ["gdiplus",
                                             "gdkpixbuf2",
                                             "pil",
                                             "quartz",
                                             "quicktime"],
                     "pyglet.gl": implementations + ["agl",
                                                     "glext_arb", "glext_nv",
                                                     "glx", "glx_info",
                                                     "glxext_arb", "glxext_mesa", "glxext_nv",
                                                     "lib_agl", "lib_glx", "lib_wgl",
                                                     "wgl", "wgl_info", "wglext_arb", "wglext_nv"],
                     "pyglet.media.drivers": ["directsound",
                                              "openal",
                                              "pulse"],
                     "pyglet.window": implementations,
                        },
                }

now = datetime.datetime.fromtimestamp(time.time())
build_data = (("Date", now.strftime("%Y/%m/%d %H:%M:%S")),
              ("pyglet version", pyglet.version))
write_build(build_data, "build.rst")

# -- SPHINX STANDARD OPTIONS ---------------------------------------------------

autosummary_generate = False

# -- General configuration -----------------------------------------------------
#
# Note that not all possible configuration values are present in this file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

inheritance_graph_attrs = dict(rankdir="LR", size='""')

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ["sphinx.ext.autodoc",
              "sphinx.ext.napoleon",
              "sphinx.ext.intersphinx",
              "sphinx.ext.inheritance_diagram",
              "sphinx.ext.todo",
              "sphinx_autodoc_typehints",
              ]

# Autodoc settings.
autodoc_member_order = "groupwise"

# Separate init from the class header as RTD theme makes it all the same color, reducing readability.
autodoc_class_signature = "separated"

# Add type hints to description and parameters in docs.
autodoc_typehints = "signature"
autodoc_typehints_format = "short"

# Configuration for sphinx_autodoc_typehints
typehints_use_signature = True
typehints_use_signature_return = True
always_use_bars_union = True

# Enable links to Python's main doc
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "pyglet"
copyright = f"2006-2008, Alex Holkner. 2008-{now.year} pyglet contributors"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
if pyglet.version.count(".") == 1:
    version = pyglet.version
else:
    version = pyglet.version.rsplit('.', 1)[0]
# The full version, including alpha/beta/rc tags.
release = pyglet.version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = 'en'

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build", "_templates", "api"]

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = False

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ["pyglet."]

# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Force custom css to allow multiline inits.
html_css_files = [
    "css/custom.css",
]
html_style = "css/custom.css"

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = ["ext/theme"]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = f"pyglet v{pyglet.version}"

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = f"pyglet v{pyglet.version} documentation "

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/logo.png"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
html_domain_indices = True

# If false, no index is generated.
html_use_index = True

# If true, the index is split into individual pages for each letter.
html_split_index = True

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "pygletdoc"

# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    ("index", "pyglet.tex", "pyglet Documentation",
     "Alex Holkner", "manual"),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ("index", "pyglet", "pyglet Documentation",
     ["Alex Holkner"], 1),
]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    ("index", "pyglet", "pyglet Documentation",
     "Alex Holkner", "pyglet", "One line description of project.",
     "Miscellaneous"),
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'

python_maximum_signature_line_length = 85

nitpicky = True


def custom_formatter(annotation, config):
    # Fixes TypeVar bounds, where the bound reference is forward referenced and class cannot be determined.
    # Defaults to a class style. Seems good enough for now since most bound things are classes.
    if isinstance(annotation, TypeVar):
        try:
            module = sphinx_autodoc_typehints.get_annotation_module(annotation)
            class_name = sphinx_autodoc_typehints.get_annotation_class_name(annotation, module)
            args = sphinx_autodoc_typehints.get_annotation_args(annotation, module, class_name)
        except ValueError:
            return str(annotation).strip("'")
        params = {k: getattr(annotation, f"__{k}__") for k in ("bound", "covariant", "contravariant")}
        params = {k: v for k, v in params.items() if v}
        if "bound" in params:
            bound_param = params["bound"]
            if isinstance(bound_param, typing.ForwardRef):
                # May be wrong but
                params["bound"] = f":py:class:`{bound_param.__forward_arg__}`"
            else:
                params["bound"] = f"{sphinx_autodoc_typehints.format_annotation(bound_param, config)}"
        args_format = f"\\(``{annotation.__name__}``{', {}' if args else ''}"
        if params:
            args_format += "".join(f", *{k} =* {v}" for k, v in params.items())
        args_format += ")"
        formatted_args = None if args else args_format
        return f":py:class:`TypeVar` {formatted_args}"
    return None


typehints_formatter = custom_formatter
